import json
import logging
import os
from typing import List, Optional, Union
from beartype import beartype
from picsellia import exceptions
from picsellia.decorators import exception_handler
from picsellia.exceptions import NoDataError, InvalidQueryError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data
from picsellia.sdk.data import MultiData
from picsellia.sdk.picture import Picture, MultiPicture
from pathlib import Path
import picsellia.pxl_multithreading as mlt
from multiprocessing import Pool, Value
import requests
import multiprocessing
from itertools import chain
from functools import partial
import tqdm
import collections

from picsellia.utils import print_line_return
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Dataset(Dao):

    def __init__(self, connexion: Connexion, id: str, name: str, version: str):
        super().__init__(connexion)
        self.id = id
        self.name = name
        self.version = version

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/dataset/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/dataset/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def fork(self, version: str, pictures: Union[MultiPicture, Picture]) -> 'Dataset':
        """Fork current dataset with a new version name.

        Will create a new dataset, with same name of current object and a new version name.
        Pictures used shall be given after being retrieved.

        Examples:
            ```python
                foo_dataset = client.fetch_dataset('datatest', 'first')
                pics = foo_dataset.list_pictures()
                bar_dataset = foo_dataset.fork('second', pics)
            ```

        Arguments:
            version (str): new version name
            pictures ((MultiPicture) or (Picture)): pictures used for the new dataset

        Returns:
            A (Dataset) with given pictures
        """
        assert version != '', 'Version name can\'t be empty'

        if isinstance(pictures, Picture):
            pictures = [pictures]

        data = json.dumps({
            'version': version,
            'picture_ids': [pic.id for pic in pictures]
        })
        r = self.connexion.post('/sdk/v1/dataset/{}/fork'.format(self.id), data=data)
        logger.info("Version {} for dataset {} created with {} assets in it".format(version, self.name, len(pictures)))
        dataset_id = r.json()["dataset"]['dataset_id']
        return Dataset(self.connexion, dataset_id, self.name, version)

    @exception_handler
    @beartype
    def _convert_response_to_picture(self, response: dict) -> Picture:
        return Picture(
            self.connexion,
            response["picture_id"],
            response["external_url"],
            response["internal_key"],
            response["width"],
            response["height"],
            response["tag"]
        )

    @exception_handler
    @beartype
    def list_pictures(self,) -> MultiPicture:
        """List all pictures of this dataset

        It will retrieve all pictures object of this dataset.
        You will then be able to manipulate them or add them to another dataset.

        Examples:
            ```python
                pictures = foo_dataset.list_pictures()
            ```

        Returns:
            A (MultiPicture) object that wraps some (Picture) that you can manipulate.
        """
        r = self.connexion.get('/sdk/v1/dataset/{}/pictures'.format(self.id)).json()
        pictures = list(map(self._convert_response_to_picture, r["pictures"]))

        if len(pictures) == 0:
            raise NoDataError("This dataset has no pictures")

        return MultiPicture(
            self.connexion,
            self.id,
            pictures
        )

    @exception_handler
    @beartype
    def get_picture(self, filename: str) -> Picture:
        """Get one picture from its filename

        Examples:
            ```python
                pic = foo_dataset.get_picture('image_1.png')
            ```
        Arguments:
            filename (str): Filename of the image in the dataset

        Returns:
            A (Picture) object that you can manipulate or download
        """
        r = self.connexion.get('/sdk/v1/dataset/{}/picture?filename={}'.format(self.id, filename)).json()
        picture = self._convert_response_to_picture(r["picture"])
        return picture

    @exception_handler
    @beartype
    def add_data(self, data: Union[Data, List[Data], MultiData]) -> None:
        """Add retrieved data from datalake to this dataset

        It will add given data into this dataset object.

        Examples:
            ```python
                data = datalake.fetch_data()[:10]
                bar_dataset.add_pictures(data)
            ```
        Arguments:
            data ((Data), List[(Data)] or (MultiData)): data to add to dataset
        """
        if isinstance(data, Data):
            data = [data]

        assert data != [], 'Please specify assets to add to dataset'

        payload = json.dumps({
            'picture_ids': [pic.id for pic in data]
        })
        self.connexion.post('/sdk/v1/dataset/{}/pictures'.format(self.id), data=payload)
        logger.info("{} assets added to Dataset {}/{}".format(len(data), self.name, self.version))

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete a dataset.

        :warning: **DANGER ZONE**: Be very careful here!

        It will remove this dataset from our database, all of its picture and annotation will be removed.

        Examples:
            ```python
                foo_dataset.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/dataset/{}'.format(self.id))
        logger.info("Dataset {} deleted".format(self.name))

    @exception_handler
    @beartype
    def update(self, type: str) -> None:
        """Update properties of a Dataset.

        Examples:
            ```python
                dataset.update(type='detection')
            ```
        """
        data = {
            "type": type
        }
        self.connexion.patch('/sdk/v1/dataset/{}'.format(self.id), json.dumps(data))
        logger.info("Dataset {} deleted".format(self.name))

    @exception_handler
    @beartype
    def download(self, target_path: str = None, nb_threads: int = 20) -> None:
        """Downloads images of a dataset.

        It will download all images from a dataset into specified folder.
        If target_path is None, it will download into ./<dataset_name>/<dataset_version>
        You can precise a number of threads to use while downloading.

        Examples:
            ```python
                foo_dataset.download('~/Downloads/dataset_pics')
            ```
        Arguments:
            target_path (str, optional): Target folder. Defaults to None.
            nb_threads (int, optional): Number of threads to use. Defaults to 20.
        """
        if target_path is not None:
            path = target_path
        else:
            path = os.path.join(self.name, self.version)

        Path(path).mkdir(parents=True, exist_ok=True)

        r = self.connexion.get('/sdk/v1/dataset/{}/download'.format(self.id)).json()
        images = r["images"]

        f = partial(mlt.mlt_download_list_external_picture, path)
        mlt.do_multiprocess_things(f, images, nb_threads)
    
        print_line_return()
        logger.info("{} images of dataset downloaded into {}".format(len(images), target_path))

    @exception_handler
    @beartype
    def get_labels(self) -> List[dict]:
        """Get all labels of a dataset

        It will retrieve a list of dictionnary representing labels.

        Examples:
            ```python
                foo_dataset.add_labels("today")
                labels = foo_dataset.get_labels()
                assert labels == [ { "name":"today", "type":"detection" }]
            ```

        Returns:
            List of label as dictionnaries
        """
        r = self.connexion.get('/sdk/v1/dataset/{}/labels'.format(self.id)).json()
        return r["labels"]

    @exception_handler
    @beartype
    def add_labels(self, labels: Union[str, List[str]]) -> None:
        """Add labels to dataset.

        Add some labels to this dataset.
        You can specify a simple label or a list of label.

        Dataset needs a type before doing this action.

        Examples:
            ```python
                foo_dataset.add_labels("today")
                foo_dataset.add_labels(["is", "a", "good", "day"])
            ```
        Arguments:
            labels (str or List[str]): label names to add
        """
        if isinstance(labels, str):
            labels = [labels]
        assert labels != [], "You shall specify some labels to add"

        data = {
            "names": labels
        }
        self.connexion.post('/sdk/v1/dataset/{}/labels'.format(self.id), data=json.dumps(data))
        logger.info("Labels {} has been added to dataset {} / {}".format(labels, self.name, self.version))

    @exception_handler
    @beartype
    def delete_labels(self, labels: Union[str, List[str]]) -> None:
        """Delete labels of a dataset.

        Remove some labels from this dataset.
        You can specify a simple label or a list of label.

        Dataset needs a type before doing this action.

        Examples:
            ```python
                foo_dataset.add_labels(["good", "day"])
                foo_dataset.remove_labels("day")
                labels = foo_dataset.get_labels()
                assert labels == [ { "name":"good", "type":"detection" }]
            ```
        Arguments:
            labels (str or List[str]): labels to remove
        """
        if isinstance(labels, str):
            labels = [labels]
        assert labels != [], "You shall specify some labels to add"

        data = {
            "names": labels
        }
        self.connexion.delete('/sdk/v1/dataset/{}/labels'.format(self.id), data=json.dumps(data))
        logger.info("Labels {} has been removed from dataset {} / {}".format(labels, self.name, self.version))

    @exception_handler
    @beartype
    def _get_annotation_page(self, limit=1, offset=0, snapshot=False):
        r = self.connexion.get(
            '/sdk/v1/dataset/{}/annotations?limit={}&offset={}&snapshot={}'.format(self.id, limit, offset, snapshot),
        ).json()
        return r

    @exception_handler
    @beartype
    def list_annotations(self, limit: int = None, offset: int = None, from_snapshot: bool = False) -> list:
        if from_snapshot:  # pragma: no cover
            annots = self._get_annotation_page(snapshot=True)
            url = annots["url"]
            object_name = annots["object_name"]
            with open(object_name, 'wb') as handler:
                response = requests.get(url, stream=True)
                total_length = response.headers.get('content-length')
                if total_length is None:
                    logger.error("Could not download {} file".format(object_name))
                else:
                    logger.debug("Downloading {}".format(object_name))
                    for data in response.iter_content(chunk_size=1024):
                        handler.write(data)
            with open(object_name, 'r') as f:
                annot_file = json.load(f)
                annotations = annot_file["annotations"]
            return annotations
        else:
            if limit is not None and offset is not None:
                annot_page = self._get_annotation_page(limit, offset)
                annotations = annot_page["annotations"]
            else:
                annotations = []
                page_size = 200
                annot_page = self._get_annotation_page(page_size, 0)
                annotations += annot_page["annotations"]
                nb_pages = annot_page["nb_pages"]
                list_pages = [i for i in range(1, nb_pages)]
                if nb_pages < 12:
                    nb_threads = nb_pages
                else:
                    try:
                        nb_threads = multiprocessing.cpu_count()
                    except Exception:
                        nb_threads = 12
                chunk_size = max(1, int(nb_pages/nb_threads))
                infos_split = list(mlt.chunks(list_pages, chunk_size))
                counter = Value('i', 0)
                result = []
                with Pool(nb_threads, initializer=mlt.init_pool_annotations,
                          initargs=(counter, nb_pages, self.connexion, self.id, page_size,)) as pool:
                    result = pool.map(mlt.dl_annotations, infos_split)
                annotations += list(chain.from_iterable(result))
            return annotations

    @exception_handler
    @beartype
    def delete_all_annotations(self) -> None:
        """Delete all annotations of this dataset

        :warning: **DANGER ZONE**: Be very careful here!

        It will remove all annotation of every pictures of this dataset

        Examples:
            ```python
                foo_dataset.delete_all_annotations()
            ```
        """
        self.connexion.delete('/sdk/v1/dataset/{}/annotations'.format(self.id)).json()

    @exception_handler
    @beartype
    def _update_stats(self, repartition: dict = None, nb_annotations: int = None, nb_objects: int = None) -> None:
        data = {
            'repartition': repartition,
            'nb_annotations': nb_annotations,
            'nb_objects': nb_objects
        }
        self.connexion.post("/sdk/v1/dataset/{}/update_stats".format(self.id), data=json.dumps(data))

    # Can't be beartyped as its used in mlt
    def _add_annotation_for_picture(self, annotation_yield,):  # pragma: no cover
        for annotations in annotation_yield:
            picture_name = annotations["external_picture_url"]
            picture = self.get_picture(picture_name)
            if "duration" in annotations.keys():
                duration = float(annotations["duration"])
            elif "time_spent" in annotations.keys():
                duration = float(annotations["time_spent"])
            nb_instances = int(annotations["nb_labels"])
            picture.add_annotation(
                data=annotations["annotations"],
                nb_instances=nb_instances,
                duration=duration)

    @exception_handler
    @beartype
    def upload_annotations_from_file(self, ann_path: str, ann_format: str,
                                     tags: Optional[list] = [], nb_jobs: Optional[int] = 5,
                                     rectangle: Optional[bool] = False):
        """Upload annotations from file
        """

        if not (os.path.isdir(ann_path) or os.path.isfile(ann_path)):
            raise InvalidQueryError('Please provide a valid directory for ann_path')

        logger.debug("Scanning categories ...")
        if ann_path is not None:
            # if ann_format == "COCO":
            #     parser = COCOParser(path=ann_path)
            #     dataset_type = parser.find_dataset_type() if not rectangle else "detection"
            #     ann_type = "rectangle" if dataset_type == "detection" else "polygon"
            #     label_names = [l["name"] for l in parser.categories]
            #     self.add_labels(label_names)

            #     logger.debug("Uploading annotations ...\n")

            #     annotations_list, repartition, nb_annotations, nb_objects = \
            #                                                     parser.generate_annotations(rectangle=rectangle)
            #     nb_threads = 20 if nb_jobs == -1 else nb_jobs
            #     pool = Pool(nb_threads)

            #     f = partial(self._upload_annotation, True, self.id)

            #     for _ in tqdm.tqdm(pool.imap_unordered(f, annotations_list), total=len(annotations_list)):
            #         pass
            #     self._update_stats(repartition=repartition, nb_annotations=nb_annotations, nb_objects=nb_objects)
            #     logger.info("Dataset Uploaded to Picsellia, you can check it on the platform.")

            # if ann_format == "PASCAL-VOC":
            #     parser = PVOCParser(path=ann_path)
            #     # TODO -> Handle segmentation cases.
            #     classes, image_infos, ann_data, repartition, nb_annotations, nb_objects = \
            #                                                   parser._generate_images_labels_annotations(tags=tags)
            #     classes = list(set(classes))
            #     label_names = [l["name"] for l in classes]
            #     self.add_labels(label_names)
            #     nb_threads = 20 if nb_jobs == -1 else nb_jobs
            #     pool = Pool(nb_threads)

            #     f = partial(self._upload_annotation, True, self.id)

            #     for _ in tqdm.tqdm(pool.imap_unordered(f, ann_data), total=len(ann_data)):
            #         pass

            #     self._update_stats(repartition=repartition, nb_annotations=nb_annotations, nb_objects=nb_objects)
            #     logger.info("Dataset Uploaded to Picsellia, you can check it on the platform.")
            if ann_format == "PICSELLIA":
                with open(ann_path, 'rb') as f:
                    dict_annotations = json.load(f)
                label_to_create = []

                repartition = collections.Counter()
                nb_annotations = 0
                nb_objects = 0

                for annotations in dict_annotations["annotations"]:
                    nb_annotations += 1
                    tmpp = []
                    for ann in annotations["annotations"]:
                        tmpp.append(ann["label"])
                        nb_objects += 1
                        tmp = [ann["type"], ann["label"]]
                        if tmp not in label_to_create:
                            label_to_create.append(tmp)
                    repartition.update(tmpp)
                label_names = [lab[1] for lab in label_to_create]
                self.add_labels(label_names)

                upload = partial(self._add_annotation_for_picture)
                nb_threads = 20 if nb_jobs == -1 else nb_jobs
                
                with Pool(nb_threads) as pool:
                    nb_threads = 20 if nb_jobs == -1 else nb_jobs
                    list_annotations = list(mlt.chunks(dict_annotations["annotations"], nb_threads))
                    for _ in tqdm.tqdm(pool.imap_unordered(upload, list_annotations), total=len(list_annotations)):
                        pass
                self._update_stats(repartition=repartition, nb_annotations=nb_annotations, nb_objects=nb_objects)
                logger.info("Dataset Uploaded to Picsellia, you can check it on the platform.")
        pass

    @exception_handler
    @beartype
    def synchronize(self, target_dir: str, do_download: bool = False) -> Union[MultiPicture, None]:
        """Synchronize this dataset with target dir by comparing pictures in target dir with pictures uploaded in dataset.

        CARE ! IT WILL JUST PRINT DIFFERENCES, AUTO DOWNLOAD/UPLOAD NOT IMPLEMENTED YET 

        Examples:
            ```python
                foo_dataset.synchronize('./foo_dataset/first')
            ```
        Arguments:
            target_dir (str): directory to synchronize against
        """
        assert os.path.isdir(target_dir), "Please select a valid directory path"
        logger.info("⌛️ Scanning Dataset Pictures ..")
        pictures = self.list_pictures()
        external_urls = set(map(lambda picture: picture.external_url, pictures))
        logger.info("🔍 Scanning Local Dataset Folder ..")
        local_filenames = set([e for e in os.listdir(target_dir) if os.path.isfile(os.path.join(target_dir, e))])


        not_uploaded = local_filenames - external_urls
        if len(not_uploaded) > 0:
            logger.info("📚 {} pictures not uploaded. You need to add data to the datalake first with :".format(len(not_uploaded)))
            logger.info("filepaths = {}".format(list(map(lambda filename: target_dir + "/" + filename, not_uploaded))))
            logger.info("list_data = client.get_datalake().upload_data(filepaths)")
            logger.info("client.get_dataset(\"{}\", \"{}\").add_data(list_data)".format(self.name, self.version))

        not_downloaded = external_urls - local_filenames 
        if len(not_downloaded) > 0:
            pictures = list(filter(lambda picture: picture.external_url in not_downloaded, pictures.picture_list))
            pics = MultiPicture(self.connexion, self.id, pictures)
            logger.info("📚 {} pictures not downloaded".format(len(not_downloaded)))
            if do_download:
                logger.info("📚 Downloading {} pictures".format(len(not_downloaded)))
                pics.download(target_dir)
            else:
                logger.info("📚 Call this method again with do_download=True if you want to download these pictures")
            return pics
        else:
            logger.info("✅ Dataset is up-to-date.")
            return None
        
        


import logging
import os
import random
from time import sleep
import zipfile
from beartype import beartype
import uuid
from picsellia import exceptions as exceptions
from picsellia.decorators import exception_handler
import sys

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    GREEN = '\033[92m'
    RED = '\033[1;31m'
    YELLOW = '\033[93m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

logger = logging.getLogger('picsellia')

try:
    download_bar_mode = os.environ["PICSELLIA_SDK_DOWNLOAD_BAR_MODE"]
except KeyError:
    download_bar_mode = "1"

@exception_handler
@beartype
def train_valid_split_obj_simple(dict_annotations, prop=0.8):
    """Perform Optimized train test split for Object Detection.
       Uses optimization to find the optimal split to have the desired repartition of instances by set.
    Arguments:
        prop (float) : Percentage of Instances used for training.
        dict_annotations (dict) : annotation from dl_annotations

    Raises:
        ResourceNotFoundError: If not annotations in the Picsell.ia Client yet."""

    if dict_annotations is None:
        raise exceptions.ResourceNotFoundError("No dict_annotations passed")

    list_im = [i for i in range(len(dict_annotations['images']))]
    random.shuffle(list_im)
    nb_im = int(prop * len(dict_annotations['images']))
    train_list = list_im[:nb_im]
    test_list = list_im[nb_im:]
    index_url = []
    for e in range(len(list_im)):
        if e in train_list:
            index_url.append(1)
        elif e in test_list:
            index_url.append(0)
    return index_url


@exception_handler
@beartype
def get_labels_repartition_obj_detection(dict_annotations, index_url):
    """Perform train test split scanning for Object Detection.
    Returns:
        cate (array[str]) : Array of the classes names
        cnt_train (array[int]) : Array of the number of object per class for the training set.
        cnt_eval (array[int]) : Array of the number of object per class for the evaluation set.

    Raises:
        ResourceNotFoundError: If not annotations in the Picsell.ia Client yet."""

    if dict_annotations is None:
        raise exceptions.ResourceNotFoundError("No dict_annotations passed")

    cate = [v["name"] for v in dict_annotations["categories"]]
    cnt_train = [0] * len(cate)
    cnt_eval = [0] * len(cate)

    for img, index in zip(dict_annotations['images'], index_url):
        internal_picture_id = img["internal_picture_id"]
        for ann in dict_annotations["annotations"]:
            if internal_picture_id == ann["internal_picture_id"]:
                for an in ann['annotations']:
                    try:
                        idx = cate.index(an['label'])
                        if index == 1:
                            cnt_train[int(idx)] += 1
                        else:
                            cnt_eval[int(idx)] += 1
                    except Exception:
                        pass
    return cnt_train, cnt_eval, cate


@exception_handler
@beartype
def zipdir(path):
    zipf = zipfile.ZipFile(path.split('.')[0] + '.zip', 'w', zipfile.ZIP_DEFLATED)
    for filepath in os.listdir(path):
        zipf.write(os.path.join(path, filepath), filepath)

        if os.path.isdir(os.path.join(path, filepath)):
            for fffpath in os.listdir(os.path.join(path, filepath)):
                zipf.write(os.path.join(path, filepath, fffpath), os.path.join(filepath, fffpath))

    zipf.close()
    return path.split('.')[0] + '.zip'


@exception_handler
@beartype
def is_uuid(string):
    try:
        uid = uuid.UUID(string, version=4)
        return str(uid) == string
    except Exception:
        return False


@exception_handler
@beartype
def check_status_code(response):
    status = int(response.status_code)
    if status in [200, 201, 202, 203]:
        try:
            response_data = response.json()
            if "success" in response_data:
                logger.debug(response_data["success"])
            if "unknown" in response_data:
                logger.warning("Some data could not be used {}".format(response_data["unknown"]))
        except Exception as e:
            logger.debug('Platform has returned (status code {}) : {}'.format(response.status_code, response.text))
        return
    elif status == 204:
        logger.debug('Resource deleted.')
        return
    elif status in [208, 400, 401, 402, 403, 404, 409, 500]:
        try:
            response_data = response.json()
            if "error" in response_data:
                message = response_data["error"]
            else:
                message = response.url
        except Exception as e:
            message = ""
            logger.warning('Platform has returned (status code {}) : {}'.format(response.status_code, response.text))
        
        if status == 208:
            raise exceptions.PicselliaError("An object has already this name in S3.")
        if status == 400:
            raise exceptions.InvalidQueryError("Invalid query : {}".format(message))
        if status == 401:
            raise exceptions.UnauthorizedError("Unauthorized : {}".format(message))
        if status == 402:
            raise exceptions.UnsufficientRessourcesError("{}".format(message))
        if status == 403:
            raise exceptions.ForbiddenError("Forbidden : {}".format(message))
        if status == 404:
            raise exceptions.ResourceNotFoundError("Resource {} is not found".format(message))
        if status == 409:
            raise exceptions.ResourceConflictError("This resouce already exists : {}".format(message))
        if status == 500:
            raise exceptions.PicselliaError("Internal server error. Please contact support. {}".format(message))
    else:
        raise Exception("Unknown error (Status code : {}), please contact support".format(response.status_code))


@exception_handler
@beartype
def generate_requirements_json(requirements_path: str):
    """Generate a json file with the requirements from the requirements.txt file

    Arguments:
        requirements_path ([str]): [absolute path to requirements.txt file]

    Raises:
        exceptions.ResourceNotFoundError: [Filepath does match]
        Exception: [Wrong requirements file]

    Returns:
        [dict]: {
            'requirements': []{
                'package': (str) package name,
                'version': (str) package version
            }
        }
    """
    js = {"requirements": []}
    try:
        with open(requirements_path, 'r') as f:
            lines = f.readlines()
    except Exception:
        raise exceptions.ResourceNotFoundError("{} does not exists".format(requirements_path))
    try:
        for line in lines:
            if line[0] != "#":
                try:
                    package, version = line.split("==")
                except Exception:
                    package, version = line, ""
                tmp = {
                    "package": package.rstrip(),
                    "version": version.rstrip()
                }
                js["requirements"].append(tmp)
    except Exception:  # pragma: no cover
        raise Exception("Malformed requirements file")
    return js


def print_next_bar(count : int, total : int):
    done = int(count * 50 / total) 

    if download_bar_mode == "0":
        return


    if download_bar_mode == "2":
        text = "[{}{}]{}%".format('=' * done, ' ' * (50 - done), done*2)
    else:
        if done == 0:
            color = bcolors.RED
            loading = "\u25CC"
        elif done < 12:
            color = bcolors.RED
            loading = "\u25CB"
        elif done < 25:
            color = bcolors.RED
            loading = "\u25D4"
        elif done < 38:
            color = bcolors.YELLOW
            loading = "\u25D1"
        elif done < 50:
            color = bcolors.YELLOW
            loading = "\u25D5"
        else:
            color = bcolors.GREEN
            loading = "\u25C9"

        text = "\r|{}{}{}{}|{} {}% {}/{}    ".format(color, '\u25AC' * done, bcolors.ENDC, ' ' * (50 - done), loading, done*2, count, total)
    sys.stdout.write(text)
    if download_bar_mode == "1":
        sys.stdout.flush()
    else:
        print_line_return()

    if done == 50 and download_bar_mode == "1":
        print_line_return()

def print_line_return():
    sys.stdout.write("\n")



import json
import logging
from typing import List
from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset import Dataset
import picsellia.utils as utils
from pathlib import Path
import os
import picsellia.exceptions as exceptions
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')

class Model(Dao):

    def __init__(self, connexion: Connexion, id: str, name: str, type: str):
        super().__init__(connexion)
        self.id = id
        self.name = name
        self.type = type

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/model/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/model/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def _convert_response_to_deployment(self, response: dict):
        from picsellia.sdk.deployment import Deployment
        return Deployment(
            self.connexion,
            response["pk"],
            response["name"],
            response["oracle_url"] if "oracle_url" in response else None,
            response["serving_url"] if "serving_url" in response else None
        )

    @exception_handler
    @beartype
    def _convert_response_to_dataset(self, response):
        dataset_id = response['dataset_id']
        dataset_name = response['dataset_name']
        dataset_version = response['version']
        return Dataset(self.connexion, dataset_id, dataset_name, dataset_version)

    @exception_handler
    @beartype
    def get_source_dataset(self) -> Dataset:
        """Get source Dataset for an Exported Modela

        Examples:
            ```python
            model = client.get_model(name="my-model")
            dataset = model.get_source_dataset()
            ```

        Returns:
            A (Dataset) that you can use and manipulate
        """
        r = self.connexion.get('/sdk/v1/model/{}/source?name={}'.format(self.id, "dataset")).json()
        return self._convert_response_to_dataset(r["dataset"])

    @exception_handler
    @beartype
    def _convert_response_to_experiment(self, response: dict):
        from picsellia.sdk.experiment import Experiment
        return Experiment(
            self.connexion,
            self.id,
            response["id"],
            response["name"],
            response["files"]
        )

    @exception_handler
    @beartype
    def _get_experiment(self, identifier: str,
                        with_artifacts: bool = False, with_logs: bool = False):
        data = {
            'with_artifacts': with_artifacts,
            'with_logs': with_logs
        }
        r = self.connexion.get(
            '/sdk/v1/model/{}/source?name={}'.format(self.id, "experiment"), data).json()

        return self._convert_response_to_experiment(r["experiment"])

    @exception_handler
    @beartype
    def get_source_experiment(self, tree: bool= False, with_artifacts: bool = False, with_logs: bool = False):
        """Retrieve an existing experiment.

        You can also to download attached files and setup according project folder tree.

        You must specify either the Experiment's name or its id.

        Examples:
            ```python
                model = client.get_model(name="my-model")
                source_experiment = model.get_source_model(with_artifacts=False, with_logs=False)
            ```
        Arguments:
            tree (bool, optional): Whether to create folder tree or not. Defaults to False.
            with_artifacts (bool, optional): Whether to download every experiment's Artifacts or not. Defaults to False.
            with_logs (bool, optional): Whether to retrieve experiment's Logs or not. Defaults to False.

        Raises:
            Exception: Experiment not found

        Returns:
            An (Experiment) object that you can manipulate
        """
        
        experiment = self._get_experiment(identifier=self.id, with_artifacts=with_artifacts, with_logs=with_logs)

        experiment._download_artifacts_and_prepare_tree(tree, with_artifacts)
        
        return experiment

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete model.

        Delete the model in Picsellia database

        Examples:
            ```python
                model.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/model/{}'.format(self.id))
        logger.info("Model {} deleted.".format(self.name))

    @exception_handler
    @beartype
    def update(self, name: str = None, tag: List[str] = None, framework: str = None, private: bool = None, 
                description: str = None, type: str = None, source_dataset: Dataset = None, source_experiment=None, 
                thumb_object_name : str = None, labels: dict = None, base_parameters: dict = None,
                notebook_link: str = None, docker_image: str=None) -> None:
        """Update model.

        Update model parameters in Picsellia database

        Examples:
            ```python
                model.update(description="Very cool model")
            ```
        """
        payload = {
            "network_name": name,
            "tag": tag,
            "framework": framework,
            "private": private,
            "description": description,
            "type": type,
            "dataset": source_dataset.id if source_dataset != None else None,
            "source_experiment": source_experiment.id if source_experiment != None else None,
            "thumb_object_name": thumb_object_name,
            "labels": labels,
            "base_parameters": base_parameters,
            "notebook_link": notebook_link,
            "image_url": docker_image
        }
        filtered = {k:v for k,v in payload.items() if v is not None}
        self.connexion.patch(
            '/sdk/v1/model/{}'.format(self.id),
            data=json.dumps(filtered)).json()
        logger.info("Model {} updated.".format(self.name))

    @exception_handler
    @beartype
    def download(self, name: str, dir_path: str = None) -> None:
        """Download file with given name stored in model.

        Examples:
            ```python
                model.download("model-latest", "./data/")
            ```
        Arguments:
            name (str): Name of the file to download
            padir_pathth (str): Directory path where file will be downloaded
        """
        r = self.connexion.get('/sdk/v1/model/{}'.format(self.id)).json()
        try:
            object_name = r["model"]["files"][name]
        except KeyError:
            raise FileNotFoundError("Could not find {} into model {}.".format(name, self.name))

        if dir_path is None:
            dir_path = './'

        path = os.path.join(dir_path, object_name)
        if self.connexion.download_some_file(False, object_name, path):
            logger.info('{} downloaded successfully'.format(object_name))
        else:
            logger.error("Could not download {} file".format(object_name))

    @exception_handler
    @beartype
    def store(self, name: str, path: str, zip: bool = False) -> None:
        """Store a file of picsellia into S3.

        Store a file into S3 database.

        Examples:
            ```python
                model.store("model-latest", "lg_test_file.pb")
            ```
        Arguments:
            name (str): Name of file
            path (str): Path of file to store
            zip (bool, optional): If true, zip directory to store it. Defaults to False.

        Raises:
            FileNotFoundError: If file not found in your path
        """
        if not os.path.exists(path):
            raise FileNotFoundError("{} not found".format(path))

        if zip:
            path = utils.zipdir(path)

        filename = os.path.split(path)[-1]
        if name == 'model-latest':
            object_name = os.path.join(self.id, '0', filename)
        else:
            object_name = os.path.join(self.id, filename)

        self.connexion.upload_file(path, object_name)

        data = {
            "filename": name,
            "object_name": object_name
        }
        self.connexion.post('/sdk/v1/model/{}/store'.format(self.id), data=json.dumps(data))
        logger.info("File {} stored for model {}".format(name, self.name))

    @exception_handler
    @beartype
    def update_thumbnail(self, thumb_path: str) -> None:
        """Updates the model thumbnail.

        Update the model thumbnail with given file.
        File size shall be less than 5Mb.

        Examples:
            ```python
                model.update_thumb("test.png")
            ```
        Arguments:
            path (str): Path of the thumbnail you want to push

        Raises:
            FileNotFoundError: If there is no file in given path
            exceptions.InvalidQueryError: If file is too large
            exceptions.PicselliaError: If an unexpected error ocurred while uploading file
        """
        if not os.path.isfile(thumb_path):
            raise FileNotFoundError("{} not found".format(thumb_path))

        filesize = Path(thumb_path).stat().st_size

        if filesize > 5*1024*1024:
            raise exceptions.InvalidQueryError("File too large limit is 5Mb")

        filename = os.path.split(thumb_path)[-1]
        object_name = os.path.join(self.id, filename)

        http_response = self.connexion.upload_file(thumb_path, object_name)
        if http_response.status_code != 204:
            raise exceptions.PicselliaError("Could not upload file : {}".format(http_response.text))

        self.update(thumb_object_name=object_name)

    @exception_handler
    @beartype
    def deploy(self, config: dict):
        """Create a (Deployment) for a model. 

        This method allows you to create a (Deployment) on Picsellia. You will then have 
        access to the monitoring dashboard and the model management part!
        If you use our serverless solution, you need to set a `min_det_threshold``
        which correspond to the confidence filter we'll apply to your predictions.
        If you want to visualize and receive 100% of the predictions, set it to 0.0

        Examples:
            Create a serverless deployment whith a min threshold of 0.2
            ```
                model = client.get_model(name="my-awesome-model")
                config = {
                    "min_det_threshold": 0.2
                }
                deployment = model.create_deployment(
                    config=config,
                    )
            ```

        Arguments:
            config (dict): Configuration file for your deployment. You can update it later.


        Returns:
            A (Deployment) that you can manipulate, connected to Picsellia
        """
        data = json.dumps({
            'config': config,
        })
        r = self.connexion.post(
            '/sdk/v1/model/{}/deploy'.format(self.id), data=data).json()
        logger.info("Deployment {} created".format(r["deployment"]["name"]))

        return self._convert_response_to_deployment(r["deployment"])

    



from beartype import beartype
from picsellia import exceptions
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao


class Organization(Dao):

    def __init__(self, connexion: Connexion, id: str, name: str):
        super().__init__(connexion)
        self.id = id
        self.name = name

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/organization/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/organization/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def get_infos(self) -> dict:
        """Return some information about this organization

        Examples:
            ```python
                org = client.get_organization()
                print(org.get_infos())
            ```

        Returns:
            A dict with id and name of the organization
        """
        return {"id": self.id, "name": self.name}



import json
import logging
from typing import List

from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
import subprocess

from picsellia.sdk.experiment import Experiment
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Run(Dao):

    def __init__(self, connexion: Connexion, id: str, config: dict, script: str,
                 script_object_name: str, requirements: list) -> None:
        super().__init__(connexion)
        self.id = id
        self.config = config
        self.script = script
        self.script_object_name = script_object_name
        self.requirements = requirements

    @exception_handler
    @beartype
    def update(self, **kwargs) -> None:
        """Update this run with the given keyword arguments.

        Examples:
            ```python
                run.update(status="running")
            ```
        """
        data = json.dumps(kwargs)
        self.connexion.patch('/sdk/v1/run/{}'.format(self.id), data=data)

    @exception_handler
    @beartype
    def end(self,) -> None:
        """End a run

        Examples:
            ```python
                run.end()
            ```
        """
        self.connexion.post('/sdk/v1/run/{}/end'.format(self.id))

    @exception_handler
    @beartype
    def download_script(self,) -> str:
        """Locally download the script from the run.

        Returns:
            Filename of the script locally downloaded
        """
        assert self.script is not None, "Run script is not yet defined"

        if self.connexion.download_some_file(False, self.script_object_name, self.script):
            logger.info('{} downloaded successfully'.format(self.script))
        else:
            logger.error("Could not download {} file".format(self.script))
        return self.script

    @exception_handler
    @beartype
    def download_data(self,) -> List[str]:
        """Download run data from the run.

        Examples:
            ```python
                run.download_data()
            ```

        Returns:
            Filenames of the data locally downloaded
        """
        assert "data" in self.config.keys(), "Run configuration is not yet defined"

        data_list = self.config["data"]
        filenames = []
        for data in data_list:
            filename = data["filename"]
            if self.connexion.download_some_file(False, data["object_name"], filename):
                filenames.append(filename)
        
        logger.info('{} downloaded successfully'.format(filenames))
        return filenames

    @exception_handler
    @beartype
    def install_requirements(self,) -> None:
        """Install requirements from the run requirements dictionnary.

        Examples:
            ```python
                run.install_requirements()
            ```
        """
        assert self.id is not None, "Please get a run first."
        req = self.requirements
        for module in req:
            name = "{}=={}".format(
                module["package"], module["version"]) if module["version"] != "" else module["package"]
            subprocess.call(['pip', 'install', name])

    @exception_handler
    @beartype
    def _convert_response_to_experiment(self, response: dict) -> Experiment:
        return Experiment(
            self.connexion,
            self.id,
            response["id"],
            response["name"],
            response["files"]
        )

    @exception_handler
    @beartype
    def get_experiment(self, ) -> Experiment:
        """Retrieve linked experiment

        Examples:
            ```python
                my_experiment = run.get_experiment()
            ```

        Returns:
            An (Experiment) object linked to this run
        """
        r = self.connexion.get('/sdk/v1/run/{}/experiment'.format(self.id)).json()
        return self._convert_response_to_experiment(r["experiment"])



from datetime import date
import json
import logging
import os
import shutil
import sys
import unittest
from unittest.mock import Mock, patch, ANY
import uuid

from requests.models import Response

from picsellia import Client, exceptions
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.deployment import ServiceMetrics
from picsellia.sdk.picture import MultiPicture, Picture

from picsellia_connexion_services import JwtServiceConnexion
from picsellia.types.classification_prediction_format import ClassificationPredictionFormat

from picsellia.types.prediction_format import PredictionFormat

unittest.defaultTestLoader.sortTestMethodsUsing = lambda *args: -1

try:
    TOKEN = os.environ["PXL_TEST_TOKEN"]
    HOST = os.environ["PXL_HOST"]
except KeyError:
    print('FATAL ERROR, you need to define env var PXL_TEST_TOKEN with api token and PXL_HOST with platform host')
    sys.exit(1)

logging.getLogger('picsellia')


class TestSDK(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.host = HOST
        cls.token = TOKEN
        return super().setUpClass()


class TestClient(TestSDK):

    def test_init_client(self):
        clt = Client(api_token=self.token, host=self.host)
        self.assertIsNotNone(clt.connexion)
        self.assertEqual(clt.connexion.host, self.host)
        self.assertEqual(clt.connexion.headers["Authorization"],"Token " + self.token)
        self.assertEqual(clt.connexion.headers["Content-type"], "application/json")
        self.assertIsNotNone(clt.organization)
        self.assertIsNotNone(clt.organization.id)
        self.assertIsNotNone(clt.organization.name)

    def test_init_environ_token(self):
        os.environ["PICSELLIA_TOKEN"] = self.token
        clt = Client(host=self.host)
        self.assertEqual(clt.connexion.headers["Authorization"],"Token " + self.token)
        self.assertEqual(clt.connexion.headers["Content-type"], "application/json")

        self.assertEqual('Client initialized for organization `{}`'.format(
            clt.organization.name), str(clt))

    def test_init_no_environ_token(self):
        os.environ["PICSELLIA_TOKEN"] = self.token
        del os.environ["PICSELLIA_TOKEN"]
        with self.assertRaises(Exception):
            Client(host=self.host)


class TestInitializedClientSDK(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.host = HOST
        cls.token = TOKEN
        cls.clt = Client(cls.token, host=cls.host)
        return super().setUpClass()


class TestOrganization(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

    def test_organization(self):
        organization = self.clt.get_organization()

        org = organization.get_infos()
        self.assertEqual(org["id"], organization.id)
        self.assertEqual(org["name"], organization.name)

    def test_get_resource_url(self):
        organization = self.clt.get_organization()
        self.assertEqual("{}/organization/{}".format(self.clt.connexion.host, organization.id),
                         organization.get_resource_url_on_platform())


class TestProject(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        project_list = cls.clt.list_projects()

        for project in project_list:
            project.delete()

    def test_create_project(self):
        project = self.clt.create_project("test_project_1")

        self.assertEqual(project.name, "test_project_1")

        project.delete()

    def test_get_datalake_platform_url(self,):
        project = self.clt.create_project("test_project_2")

        base_url = self.clt.connexion.host
        self.assertEqual("{}/project/{}".format(base_url, project.id),
                         project.get_resource_url_on_platform())

    def test_list_project(self):
        project = self.clt.create_project("test_project_3")

        project_list = self.clt.list_projects()

        self.assertEqual(len(project_list), 1)
        self.assertEqual(project_list[0].name, "test_project_3")

        project.delete()

    def test_update_project(self):
        project = self.clt.create_project("test_project_4")

        project.update(description="test project description")

        project.delete()

    def test_get_project_by_name(self):
        project = self.clt.create_project("test_project_5")

        project_retr = self.clt.get_project("test_project_5")
        self.assertEqual(project.name, project_retr.name)

        project.delete()

    def test_get_project_by_id(self):
        project = self.clt.create_project("test_project_6")

        project_retr = self.clt.get_project_by_id(project.id)
        self.assertEqual(project.id, project_retr.id)
        self.assertEqual(project.name, project_retr.name)

        project.delete()

    def test_list_workers(self):
        project = self.clt.create_project("test_project_7")

        workers = project.list_workers()
        self.assertTrue(len(workers) > 0)
        self.assertIsNotNone(workers[0].get_infos()["username"])

        project.delete()


class TestDatalake(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        try:
            datalist = datalake.list_data()
            for one_data in datalist:
                one_data.delete()
        except exceptions.NoDataError:
            datalist = []

    def test_get_datalake(self,):
        datalake = self.clt.get_datalake()
        self.assertIsNotNone(datalake)

    def test_get_datalake_platform_url(self,):
        datalake = self.clt.get_datalake()
        base_url = self.clt.connexion.host
        self.assertEqual("{}/datalake/{}".format(base_url, datalake.id),
                         datalake.get_resource_url_on_platform())

    def test_upload_one_data(self,):
        self.clear_data()

        datalake = self.clt.get_datalake()
        data_list_input = datalake.upload_data(
            'tests/files/test.png', ["tag1"])
        data_list = datalake.fetch_data(tags=["tag1"])
        self.assertIsNotNone(data_list)
        self.assertEqual(len(data_list.data_list), 1)
        self.assertEqual(data_list[0].id, data_list_input.id)

    def test_upload_zero_data(self,):
        self.clear_data()

        datalake = self.clt.get_datalake()
        with self.assertRaises(exceptions.PicselliaError):
            datalake.upload_data('tests/files/unknown.png', ["tag1"])

    def test_upload_one_unknown_data_but_other_good(self,):
        self.clear_data()

        datalake = self.clt.get_datalake()
        data_list_input = datalake.upload_data(
            ['tests/files/test.png', 'unknown'], ["tag2"])

        data_list = datalake.list_data()
        self.assertIsNotNone(data_list)
        self.assertEqual(len(data_list.data_list), 1)
        self.assertEqual(data_list[0].id, data_list_input.id)

    def test_upload_multiple_data(self,):
        self.clear_data()

        datalake = self.clt.get_datalake()
        files = os.listdir('tests/files/test_image_dir')
        files = ['tests/files/test_image_dir/' + fle for fle in files]
        datalake.upload_data(files, ["tag2"])

        data_list = datalake.fetch_data(tags=["tag2"])
        self.assertEqual(len(data_list.data_list), len(files))

    def clear_data(self):
        datalake = self.clt.get_datalake()
        try:
            datalist = datalake.list_data()
            for one_data in datalist:
                one_data.delete()
        except exceptions.NoDataError:
            datalist = []


class TestData(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        try:
            datalist = datalake.list_data()
        except exceptions.NoDataError:
            datalist = []
        for one_data in datalist:
            one_data.delete()

        files = os.listdir('tests/files/test_image_dir')
        cls.files = ['tests/files/test_image_dir/' + fle for fle in files]
        cls.datalake = cls.clt.get_datalake()
        cls.datalake.upload_data(cls.files, ["tag3"])

    def test_fetch_data(self):
        data_list = self.datalake.fetch_data()
        self.assertEqual(len(data_list.data_list), 6)

        with self.assertRaises(exceptions.NoDataError):
            self.datalake.fetch_data(tags=["tag1"])

        data_list = self.datalake.fetch_data(tags=["tag3"], quantity=100)
        self.assertEqual(len(data_list.data_list), len(self.files))

        data_list = self.datalake.fetch_data(tags=["tag3"], quantity=2)
        self.assertEqual(len(data_list.data_list), 2)

        data_list = self.datalake.fetch_data(tags="tag3", quantity=2)
        self.assertEqual(len(data_list.data_list), 2)

    def test_one_data_tags(self):
        one_data = self.datalake.fetch_data(quantity=1, tags=["tag3"])[0]

        one_data.add_tags("a")
        one_data.add_tags(["b", "c"])

        with self.assertRaises(AssertionError):
            one_data.add_tags([])

        tags = one_data.get_tags()
        self.assertIn("a", tags)
        self.assertIn("b", tags)
        self.assertIn("c", tags)

        one_data.remove_tags("a")
        one_data.remove_tags(["b", "c"])

        with self.assertRaises(AssertionError):
            one_data.remove_tags([])

        tags = one_data.get_tags()
        self.assertEqual(tags, ["tag3"])

    def test_add_tags_on_multiple(self):
        multidata = self.datalake.fetch_data(quantity=2)

        multidata.add_tags("a")
        multidata.add_tags(["b", "c"])

        with self.assertRaises(AssertionError):
            multidata.add_tags([])

        tags = multidata[0].get_tags()
        self.assertIn("a", tags)
        self.assertIn("b", tags)
        self.assertIn("c", tags)

        tags = multidata[1].get_tags()
        self.assertIn("a", tags)
        self.assertIn("b", tags)
        self.assertIn("c", tags)

        multidata.remove_tags("a")
        multidata.remove_tags(["b", "c"])

        with self.assertRaises(AssertionError):
            multidata.remove_tags([])

        tags = multidata[0].get_tags()
        self.assertEqual(tags, ["tag3"])

        tags = multidata[1].get_tags()
        self.assertEqual(tags, ["tag3"])

    def test_delete_data(self):
        one_data = self.datalake.fetch_data(quantity=1)[0]
        one_data.add_tags('T1')
        one_data.delete()

        with self.assertRaises(exceptions.NoDataError):
            self.datalake.fetch_data(tags=['T1'])

    def test_delete_multidata(self):
        multidata = self.datalake.fetch_data(quantity=2)
        multidata.add_tags('T1')
        multidata.delete()

        with self.assertRaises(exceptions.NoDataError):
            self.datalake.fetch_data(tags=['T1'])

    def test_checks_on_multidata(self):
        with self.assertRaises(exceptions.NoDataError):
            MultiData(self.datalake.connexion, self.datalake.id, [])

        with self.assertRaises(TypeError):
            MultiData(self.datalake.connexion, self.datalake.id, [
                      Picture(self.datalake.connexion, "1", "2", "3")])

        one_data = self.datalake.fetch_data(quantity=1)[0]
        multidata = self.datalake.fetch_data(quantity=2)
        multidata[1] = one_data

        with self.assertRaises(IndexError):
            multidata[2] = one_data

        self.assertEqual(len(multidata), 2)
        del multidata[0]
        self.assertEqual(len(multidata), 1)
        with self.assertRaises(exceptions.NoDataError):
            del multidata[1]

        multidata = self.datalake.fetch_data(quantity=3)[:2]
        self.assertEqual(len(multidata), 2)


    def test_download_data(self):
        multidata = self.datalake.fetch_data(quantity=3)

        data = multidata[0]
        data.download()
        path = "./{}".format(data.external_url)
        self.assertTrue(os.path.exists(path))
        os.remove(path)
        
        data.download('../downloads')
        path = "../downloads/{}".format(data.external_url)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

        multidata.download()
        for d in multidata:
            path = "./{}".format(d.external_url)
            self.assertTrue(os.path.exists(path))
            os.remove(path)

        multidata.download('../downloads')
        for d in multidata:
            path = "../downloads/{}".format(d.external_url)
            self.assertTrue(os.path.exists(path))
            os.remove(path)

        data = multidata[0]
        data.download()
        path = "./{}".format(data.external_url)
        self.assertTrue(os.path.exists(path))
        data.download() # Override file
        self.assertTrue(os.path.exists(path))
        os.remove(path)

class TestDataset(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        try:
            datasets = cls.clt.list_datasets()
        except exceptions.ResourceNotFoundError:
            datasets = []

        for dataset in datasets:
            dataset.delete()

        datalake = cls.clt.get_datalake()

        files = os.listdir('tests/files/test_image_dir')
        files = ['tests/files/test_image_dir/' + fle for fle in files]
        datalake = cls.clt.get_datalake()
        datalake.upload_data(files, ["tag3"])
        data_list = datalake.fetch_data(quantity=2)

        cls.dataset = cls.clt.create_dataset("datatest", data_list)
        cls.dataset.update('detection')

    def test_create_dataset_with_only_one_image(self,):
        datalake = self.clt.get_datalake()

        data_list = datalake.fetch_data(quantity=1)[0]

        dataset = self.clt.create_dataset("datatest_create", data_list)
        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.name, "datatest_create")
        self.assertEqual(dataset.version, "first")
        dataset.delete()

    def test_create_dataset_from_a_list(self,):
        datalake = self.clt.get_datalake()

        data_list = datalake.fetch_data(quantity=1)[:5].data_list

        dataset = self.clt.create_dataset("datatest_create", data_list)
        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.name, "datatest_create")
        self.assertEqual(dataset.version, "first")
        dataset.delete()

    def test_get_dataset_platform_url(self,):
        datalake = self.clt.get_datalake()

        data_list = datalake.fetch_data(quantity=1)[0]

        dataset = self.clt.create_dataset("datatest_get_url", data_list)
        base_url = self.clt.connexion.host
        self.assertEqual("{}/dataset/{}".format(base_url, dataset.id),
                         dataset.get_resource_url_on_platform())

    def test_create_dataset(self,):
        datalake = self.clt.get_datalake()

        data_list = datalake.fetch_data(quantity=1)

        dataset = self.clt.create_dataset("datatest_create", data_list)
        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.name, "datatest_create")
        self.assertEqual(dataset.version, "first")
        dataset.delete()

    def test_get_dataset_by_name(self,):
        dataset = self.clt.get_dataset("datatest", "first")

        self.assertIsNotNone(dataset)
        self.assertEqual(dataset.name, self.dataset.name)
        self.assertEqual(dataset.version, self.dataset.version)
        self.assertEqual(dataset.id, self.dataset.id)

    def test_get_dataset_by_id(self,):
        dataset = self.clt.get_dataset("datatest", "first")
        dataset_from_id = self.clt.get_dataset_by_id(dataset.id)

        self.assertIsNotNone(dataset)
        self.assertEqual(dataset_from_id.name, self.dataset.name)
        self.assertEqual(dataset_from_id.version, self.dataset.version)
        self.assertEqual(dataset_from_id.id, self.dataset.id)

    def test_search_dataset(self,):
        data_list = self.clt.datalake.fetch_data(quantity=1)
        dataset1 = self.clt.create_dataset("d1", data_list, 'v1')
        dataset2 = self.clt.create_dataset("d2", data_list, 'v1')

        pics = dataset1.list_pictures()[:1]
        dataset3 = dataset1.fork("v2", pics)

        self.assertIsNotNone(dataset1)
        self.assertIsNotNone(dataset2)
        self.assertIsNotNone(dataset3)

        ids_1 = [d.id for d in self.clt.search_datasets(name="d1")]
        self.assertIn(dataset1.id, ids_1)
        self.assertIn(dataset3.id, ids_1)
        self.assertNotIn(dataset2.id, ids_1)

        ids_2 = [d.id for d in self.clt.search_datasets(version="v1")]
        self.assertIn(dataset1.id, ids_2)
        self.assertIn(dataset2.id, ids_2)
        self.assertNotIn(dataset3.id, ids_2)

        ids_3 = [d.id for d in self.clt.search_datasets(
            name="d2", version="v1")]
        self.assertIn(dataset2.id, ids_3)
        self.assertNotIn(dataset1.id, ids_3)
        self.assertNotIn(dataset3.id, ids_3)

    def test_list_pictures_without_data(self,):
        data_list = self.clt.datalake.fetch_data(quantity=1)
        dataset1 = self.clt.create_dataset("test_dataset_without_pic", data_list)

        pics = dataset1.list_pictures()
        pics.delete()

        self.assertIsNotNone(dataset1)

        with self.assertRaises(exceptions.NoDataError):
            dataset1.list_pictures()

    def test_fork_dataset_with_one_pic(self,):
        pics = self.dataset.list_pictures()[0]

        new_dataset = self.dataset.fork("second", pics)
        self.assertIsNotNone(new_dataset)
        self.assertEqual(new_dataset.name, self.dataset.name)
        self.assertEqual(new_dataset.version, "second")
        self.assertNotEqual(new_dataset.id, self.dataset.id)

        new_dataset.delete()

    def test_fork_dataset(self,):
        pics = self.dataset.list_pictures()[:1]

        new_dataset = self.dataset.fork("second", pics)
        self.assertIsNotNone(new_dataset)
        self.assertEqual(new_dataset.name, self.dataset.name)
        self.assertEqual(new_dataset.version, "second")
        self.assertNotEqual(new_dataset.id, self.dataset.id)

        new_dataset.delete()

    def test_add_data_to_dataset(self,):
        datalake = self.clt.get_datalake()
        data_list = datalake.fetch_data(quantity=6)
        dataset2 = self.clt.create_dataset("datatest2", data_list[2:4])

        dataset2.add_data(data_list[0])
        dataset2.add_data([data_list[0], data_list[1]])
        dataset2.add_data(data_list[0:2])
        dataset2.add_data(data_list)
        dataset2.delete()

    def test_add_labels(self,):
        self.dataset.add_labels("car")

        labels = self.dataset.get_labels()
        self.assertEqual(labels[0]["name"], "car")

        self.dataset.delete_labels("car")

        labels = self.dataset.get_labels()
        self.assertEqual(len(labels), 0)

    def test_download_dataset(self):
        self.dataset.download()
        self.assertTrue(os.path.exists("datatest"))
        shutil.rmtree("datatest")

        self.dataset.download("tests/files/test_dl_dataset")
        self.assertTrue(os.path.exists("tests/files/test_dl_dataset"))
        shutil.rmtree("tests/files/test_dl_dataset")

    def test_upload_annotations_from_file_with_unknown_file(self,):
        with self.assertRaises(exceptions.InvalidQueryError):
            self.dataset.upload_annotations_from_file(
                ann_path="tests/files/not_a_file.json", ann_format="PICSELLIA")

    def test_upload_annotations_from_file_unknown_without_images(self,):
        with self.assertRaises(exceptions.InvalidQueryError):
            self.dataset.upload_annotations_from_file(
                ann_path="tests/files/not_a_file.json", tags=["test_push"], ann_format="PICSELLIA")

    def test_upload_annotations_from_file_unparsable_without_images(self,):
        # missing images 'key' in json
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.dataset.upload_annotations_from_file(
                ann_path="tests/files/test_annotations_wrong.json", tags=["test_push"], ann_format="PICSELLIA")

    def test_upload_annotations_from_file_list_delete(self,):
        datalake = self.clt.get_datalake()
        file_list = os.listdir("tests/files/test_push_dataset_dir")
        filepaths = [os.path.join("tests/files/test_push_dataset_dir", f)
                     for f in file_list]
        data_list = datalake.upload_data(filepaths)

        dataset = self.clt.create_dataset("test_push_dataset", data_list)
        dataset.update('detection')
        dataset.upload_annotations_from_file(
            ann_path="tests/files/test_annotations_rectangle.json", ann_format="PICSELLIA")
        annotations = dataset.list_annotations()  # test list all annotations
        self.assertGreater(len(annotations), 0)
        annotations = dataset.list_annotations(1, 0)  # test list with offset and limit

        # TODO annotations = dataset.list_annotations(1, 0, from_snapshot=True)

        dataset.delete_all_annotations()

    def test_create_dataset(self,):
        files = os.listdir('tests/files/test_a_lot')
        files = ['tests/files/test_a_lot/' + fle for fle in files]
        datalake = self.clt.get_datalake()
        datalake.upload_data(files, ["tag_a_lot"])
        data_list = datalake.fetch_data(quantity=25)

        dataset = self.clt.create_dataset("datatest_a_lot_1", data_list[0:10])
        self.assertEqual(10, len(dataset.list_pictures()))
        dataset.delete()
        dataset = self.clt.create_dataset("datatest_a_lot_2", data_list[0:20])
        self.assertEqual(20, len(dataset.list_pictures()))
        dataset.delete()
        dataset = self.clt.create_dataset("datatest_a_lot_3", data_list[0:25])
        self.assertEqual(25, len(dataset.list_pictures()))
        dataset.delete()


class TestPicture(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        files = os.listdir('tests/files/test_push_dataset_dir')
        files = ['tests/files/test_push_dataset_dir/' + fle for fle in files]
        datalake = cls.clt.get_datalake()
        datalake.upload_data(files, ["test_pics"])
        data_list = datalake.fetch_data(tags="test_pics")
        cls.dataset = cls.clt.create_dataset("datatest_picture", data_list)
        cls.dataset.update('detection')

    def test_checks_on_multidata(self):
        with self.assertRaises(exceptions.NoDataError):
            MultiPicture(self.dataset.connexion, self.dataset.id, [])

        with self.assertRaises(TypeError):
            MultiPicture(self.dataset.connexion, self.dataset.id, [
                         Data(self.dataset.connexion, "1", "2", "3", "4")])

        one_pic = self.dataset.list_pictures()[0]
        self.assertEqual(one_pic.tags, ["test_pics"])
        self.assertTrue(one_pic.width > 0)
        self.assertTrue(one_pic.height > 0)
        multipic = self.dataset.list_pictures()[:2]
        self.assertTrue(isinstance(one_pic, Picture))
        self.assertTrue(len(multipic), 2)

        multipic[1] = one_pic

        with self.assertRaises(IndexError):
            multipic[2] = one_pic

        self.assertEqual(len(multipic), 2)
        del multipic[0]
        self.assertEqual(len(multipic), 1)
        with self.assertRaises(exceptions.NoDataError):
            del multipic[1]

        pics = self.dataset.list_pictures()
        for pic in pics:
            pic.delete()



    def test_download_pictures(self):
        multipics = self.dataset.list_pictures()

        pic = multipics[0]
        pic.download()
        path = "./{}".format(pic.external_url)
        self.assertTrue(os.path.exists(path))
        os.remove(path)
        
        pic.download('../downloads')
        path = "../downloads/{}".format(pic.external_url)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

        multipics.download()
        for d in multipics:
            path = "./{}".format(d.external_url)
            self.assertTrue(os.path.exists(path))
            os.remove(path)

        multipics.download('../downloads')
        for d in multipics:
            path = "../downloads/{}".format(d.external_url)
            self.assertTrue(os.path.exists(path))
            os.remove(path)

        pic = multipics[0]
        pic.download()
        path = "./{}".format(pic.external_url)
        self.assertTrue(os.path.exists(path))
        pic.download() # Override file
        self.assertTrue(os.path.exists(path))
        os.remove(path)


    def test_syncronize(self):
        multipics = self.dataset.list_pictures()
        self.assertTrue(len(multipics) > 0)

        target_path = "./test_sync/"
        os.mkdir(target_path)
        diff = self.dataset.synchronize(target_path)

        # Empty dir
        self.assertEqual(len(diff), len(multipics))

        # No download by default
        self.assertEqual(len(os.listdir(target_path)), 0)
        
        self.dataset.synchronize(target_path, do_download=True)
        files = os.listdir(target_path)
        self.assertEqual(len(files), len(multipics))

        diff = self.dataset.synchronize(target_path)
        self.assertIsNone(diff)

        os.remove(target_path + "/" + files[0])
        diff = self.dataset.synchronize(target_path)
        self.assertEqual(len(diff), 1)

        diff = self.dataset.synchronize(target_path, do_download=True)
        files = os.listdir(target_path)
        self.assertEqual(len(files), len(multipics))

        shutil.rmtree(target_path)


class TestPictureAnnotation(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        files = os.listdir('tests/files/test_push_dataset_dir')
        files = ['tests/files/test_push_dataset_dir/' + fle for fle in files]
        datalake = cls.clt.get_datalake()
        datalake.upload_data(files, ["test_pics"])
        data_list = datalake.fetch_data(tags="test_pics")
        cls.dataset = cls.clt.create_dataset(
            "datatest_picture_annotations", data_list)
        cls.dataset.update('detection')

    def test_add_list_delete_annotation(self):
        filename = "image_11.jpg"
        self.dataset.list_pictures()
        picture = self.dataset.get_picture(filename)
        with open("tests/files/test_annotations_rectangle.json", 'r') as f:
            annot_dict = json.load(f)
            annotation_data = annot_dict["annotations"][1]["annotations"]
            picture.add_annotation(data=annotation_data, creation_date=date.today(),
                                   duration=10.0, reviewed=False, is_accepted=True, is_skipped=True, nb_instances=10)
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 1)
        picture.delete_annotations()
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 0)

        with open("tests/files/test_annotations_classif.json", 'r') as f:
            annot_dict = json.load(f)
            annotation_data = annot_dict["annotations"][1]["annotations"]
            picture.add_annotation(data=annotation_data)
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 1)
        picture.delete_annotations()
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 0)

        with open("tests/files/test_annotations_polygon.json", 'r') as f:
            annot_dict = json.load(f)
            annotation_data = annot_dict["annotations"][0]["annotations"]
            picture.add_annotation(data=annotation_data)
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 1)
        picture.delete_annotations()
        annots = picture.list_annotations()
        self.assertEqual(len(annots), 0)

        picture = self.dataset.get_picture("image_11.jpg")

        with self.assertRaises(TypeError):
            picture.add_annotation(data={})

        with self.assertRaises(exceptions.TyperError):
            picture.add_annotation(data=[])

        with self.assertRaises(TypeError):
            picture.add_annotation(data=[[]])

        with self.assertRaises(exceptions.TyperError):
            # Dict empty in 2
            picture.add_annotation(data=[{}])

        with self.assertRaises(exceptions.TyperError):
            # Type not present
            picture.add_annotation(data=[{"label": "car"}])

        with self.assertRaises(exceptions.TyperError):
            # Type unknown
            picture.add_annotation(data=[{"type": "unknown"}])

        with self.assertRaises(exceptions.TyperError):
            # Label not present
            picture.add_annotation(data=[{"type": "classification"}])

        with self.assertRaises(exceptions.TyperError):
            # no rectangle
            data = [{"type": "rectangle", "label": "car"}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no height
            data = [{"type": "rectangle", "label": "car",
                     "rectangle": {"top": 1, "left": 1, "width": 1}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no top
            data = [{"type": "rectangle", "label": "car",
                     "rectangle": {"height": 1, "left": 1, "width": 1}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no width
            data = [{"type": "rectangle", "label": "car",
                     "rectangle": {"left": 1, "top": 1, "height": 1}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no left
            data = [{"type": "rectangle", "label": "car",
                     "rectangle": {"width": 1, "top": 1, "height": 1}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no polygon
            data = [{"type": "polygon", "label": "car"}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # polygon bad type
            data = [{"type": "polygon", "label": "car", "polygon":  "HEY"}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no geometry
            data = [{"type": "polygon", "label": "car", "polygon": {}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # bad geo
            data = [{"type": "polygon", "label": "car",
                     "polygon":  {"geometry": "YAA"}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # empty geo
            data = [{"type": "polygon", "label": "car",
                     "polygon":  {"geometry": []}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # tiny geo
            data = [{"type": "polygon", "label": "car", "polygon":  {
                "geometry": [{"x": 1, "y": 2}, {"x": 3, "y": 4}]}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # bad coord
            data = [{"type": "polygon", "label": "car", "polygon":  {
                "geometry": [{"x": 1, "y": 2}, {"x": 3, "y": 4}, "HEYO"]}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no x
            data = [{"type": "polygon", "label": "car", "polygon":  {
                "geometry": [{"x": 1, "y": 2}, {"x": 3, "y": 4}, {"y": 4}]}}]
            picture.add_annotation(data=data)

        with self.assertRaises(exceptions.TyperError):
            # no y
            data = [{"type": "polygon", "label": "car", "polygon":  {
                "geometry": [{"x": 1, "y": 2}, {"x": 3, "y": 4}, {"x": 4}]}}]
            picture.add_annotation(data=data)

        pics = self.dataset.list_pictures()
        pics.delete()


class TestModel(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        models = cls.clt.list_models()
        for model in models:
            model.delete()
        try:
            datalist = datalake.list_data()
        except exceptions.NoDataError:
            datalist = []
        for one_data in datalist:
            one_data.delete()
        project_list = cls.clt.list_projects()
        for project in project_list:
            project.delete()

        files = os.listdir('tests/files/test_image_dir')
        cls.files = ['tests/files/test_image_dir/' + fle for fle in files]
        cls.datalake = cls.clt.get_datalake()
        cls.datalake.upload_data(cls.files, ["tag3"])
        data_list = datalake.fetch_data(quantity=2)
        cls.dataset = cls.clt.create_dataset("datatest", data_list)
        cls.dataset.update('detection')
        cls.project = cls.clt.create_project("test_project_1")
        cls.experiment = cls.project.create_experiment(
            name="test_experiment_1")
        cls.parameters = {
            "param1": 1
        }
        cls.experiment.log('parameters', data=cls.parameters, type="table")
        cls.model = cls.clt.create_model("test_model", "detection")
        cls.model_with_sources = cls.clt.create_model("test_model_with_source", "detection")
        cls.model_with_sources.update(
            source_dataset=cls.dataset,
        )
        cls.model_with_sources.update(
            source_experiment=cls.experiment
        )

        cls.model.store('model-latest', 'tests/files/test_dir', True)

    def test_get_model_platform_url(self,):
        base_url = self.clt.connexion.host
        self.assertEqual("{}/model/{}".format(base_url, self.model.id),
                         self.model.get_resource_url_on_platform())

    def test_create_model(self):
        model = self.clt.create_model("model_1", "classification")

        self.assertEqual(model.name, "model_1")
        self.assertEqual(model.type, "classification")

        model.delete()

    def test_create_model_same_name_error(self):
        with self.assertRaises(exceptions.ResourceConflictError):
            self.clt.create_model(self.model.name, "classification")

    def test_get_model_by_id(self):
        model = self.clt.create_model("model_1", "classification")

        model_retr = self.clt.get_model_by_id(model.id)
        self.assertEqual(model.id, model_retr.id)
        self.assertEqual(model.name, model_retr.name)
        self.assertEqual(model.type, model_retr.type)

        model.delete()

    def test_get_model_by_id_not_found(self):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.clt.get_model_by_id(
                "014630b4-691a-41f0-a48d-cf2e1f9487f0")

    def test_get_model_by_name(self):
        model_retr = self.clt.get_model("test_model")

        self.assertEqual(self.model.id, model_retr.id)
        self.assertEqual(self.model.name, model_retr.name)
        self.assertEqual(self.model.type, model_retr.type)

    def test_get_model_by_name_not_found(self):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.clt.get_model("unknown")

    def test_list_models(self):
        model2 = self.clt.create_model("model_2", "classification")
        model3 = self.clt.get_model("test_model_with_source")
        models = self.clt.list_models()

        self.assertIsNotNone(models)
        self.assertEqual(len(models), 3)

        # sort to ensure order
        names = sorted([self.model.name, model2.name, model3.name])
        names_retr = sorted([model.name for model in models])

        ids = sorted([self.model.id, model2.id, model3.id])
        ids_retr = sorted([model.id for model in models])

        self.assertListEqual(names, names_retr)
        self.assertListEqual(ids, ids_retr)

        model2.delete()

    def test_update_model(self):
        self.model.update(labels={'0': 'car', '1': 'bird'})
        self.model.update(description="Yes yes yes", tag=["test_tag"])

    def test_model_store(self,):
        with self.assertRaises(FileNotFoundError):
            self.model.store('config', 'unknown')

        self.model.store('config', 'tests/files/sm_test_file.config')
        # test file already stored
        self.model.store('config', 'tests/files/sm_test_file.config')
        path_dwl = './tests/files/test_dir.zip'
        self.assertTrue(os.path.exists(path_dwl))
        os.remove(path_dwl)

        self.model.store('large_file', 'tests/files/lg_test_file.pb')

        with self.assertRaises(FileNotFoundError):
            self.model.store('unknown', 'unknown')

        self.assertIsNotNone(self.model.id)
        self.model.download('config')
        path_dwl = './{}'.format(self.model.id)
        self.assertTrue(os.path.exists(path_dwl))
        shutil.rmtree(path_dwl)

        self.model.download('model-latest', './tests/files/')
        path_dwl = './tests/files/{}'.format(self.model.id)
        self.assertTrue(os.path.exists(path_dwl))
        shutil.rmtree(path_dwl)

        with self.assertRaises(FileNotFoundError):
            self.model.download('unknown')

    def test_update_thumb(self,):
        self.model.update_thumbnail("tests/files/test.png")

        with self.assertRaises(FileNotFoundError):
            self.model.update_thumbnail("unknown")

        with self.assertRaises(exceptions.InvalidQueryError):
            self.model.update_thumbnail("tests/files/lg_test_file.pb")

    def test_deploy_mocked_pc(self,):
        config = {
            "min_det_threshold": 0.2
        }
        with patch('sdk.model.Connexion.post') as mock_connexion:
            response = Mock(spec=Response)
            response.json.return_value = {"deployment" : {"pk": "iddeployment", "name": "name-test"}}
            response.status_code = 201
            mock_connexion.return_value = response

            deployment = self.model.deploy(
                config=config
            )
        
        with patch('sdk.model.Connexion.get') as mock_connexion:
            response = Mock(spec=Response)
            response.json.return_value = {"deployment" : {"pk": "iddeployment", "name": "name-test"}}
            response.status_code = 200
            mock_connexion.return_value = response

            fetched_deployment = self.clt.get_deployment(deployment.name)
        
        self.assertEqual(deployment.name, fetched_deployment.name)

    def test_deploy(self,):
        config = {
            "min_det_threshold": 0.2
        }

        with patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion)):
            deployment = self.model.deploy(config=config)
            fetched_deployment = self.clt.get_deployment(deployment.name)
        
        self.assertEqual(deployment.name, fetched_deployment.name)

    def test_deploy_wrong_config_key(self,):
        config = {
            "min_det_threshold_error": 0.2
        }
        with self.assertRaises(exceptions.InvalidQueryError):
            deployment = self.model.deploy(
                config=config
            )

    def test_deploy_wrong_config_type(self,):
        config = {
            "min_det_threshold": "0.2"
        }
        with self.assertRaises(exceptions.InvalidQueryError):
            deployment = self.model.deploy(
                config=config
            )

    def test_deploy_wrong_config_interval(self,):
        config = {
            "min_det_threshold": 20.0
        }
        with self.assertRaises(exceptions.InvalidQueryError):
            deployment = self.model.deploy(
                config=config
            )

    def test_get_source_dataset(self,):
        source_dataset = self.model_with_sources.get_source_dataset()
        self.assertEqual(source_dataset.name, self.dataset.name)
        self.assertEqual(source_dataset.version, self.dataset.version)
        self.assertEqual(source_dataset.id, self.dataset.id)

    def test_get_source_dataset_no_dataset(self,):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.model.get_source_dataset()

    def test_get_source_experiment(self,):
        source_experiment = self.model_with_sources.get_source_experiment(
            False, False, False
        )
        self.assertEqual(source_experiment.name, self.experiment.name)
        self.assertEqual(source_experiment.id, self.experiment.id)
    
    def test_get_source_experiment_no_experiment(self,):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.model.get_source_experiment()
    

        
        

class TestExperimentCRUD(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.project = cls.clt.create_project("test_project_experiment_crud")
        cls.experiment = cls.project.create_experiment(
            name="test_experiment_1")
        cls.exp_id = cls.experiment.id

    def test_get_experiment_by_name(self):
        experiment = self.project.get_experiment(name="test_experiment_1")
        self.assertEqual(self.exp_id, experiment.id)

    def test_get_experiment_by_id(self):
        experiment = self.project.get_experiment(id=self.exp_id)
        self.assertEqual(self.exp_id, experiment.id)

    def test_get_experiment_without_identifier_error(self):
        with self.assertRaises(exceptions.PicselliaError):
            self.project.get_experiment()

    def test_list_experiments(self):
        experiments = self.project.list_experiments()
        self.assertEqual(len(experiments), 1)
        self.assertEqual(self.exp_id, experiments[0].id)

    def test_update_experiment(self):
        self.experiment.update(description="test")

    def test_delete_experiment(self):
        experiment = self.project.create_experiment(
            name="test_experiment_delete")

        experiment.delete()
        with self.assertRaises(exceptions.ResourceNotFoundError):
            experiment = self.project.get_experiment(
                name="test_experiment_delete")

    def test_create_experiment_from_another(self):
        exp1 = self.project.create_experiment(name="test_experiment_3")
        exp2 = self.project.create_experiment(name="test_experiment_4", previous=exp1)

        exp_retr = self.project.get_experiment("test_experiment_4")
        self.assertEqual(exp2.id, exp_retr.id)

        exp2.delete()
        exp1.delete()

    def test_create_experiment_from_model_source(self):
        model = self.clt.create_model("model2", "classification")
        exp = self.project.create_experiment(name="test_experiment_5", source=model)

        exp_retr = self.project.get_experiment("test_experiment_5")
        self.assertEqual(exp.id, exp_retr.id)

        exp.delete()
        model.delete()

    def test_delete_all_experiments(self):
        self.project.create_experiment(name="test_experiment_2")

        self.project.delete_all_experiments()
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.project.get_experiment(name="test_experiment_1")
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.project.get_experiment(name="test_experiment_2")


class TestExperiment(TestInitializedClientSDK):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.project = cls.clt.create_project("test_project_experiment")
        cls.experiment = cls.project.create_experiment(
            name="test_experiment_1")
        cls.exp_id = cls.experiment.id
        files = os.listdir('tests/files/test_push_dataset_dir')
        files = ['tests/files/test_push_dataset_dir/' + fle for fle in files]
        datalake = cls.clt.get_datalake()
        data_list = datalake.upload_data(files, ["tag_exp"])
        dataset = cls.clt.create_dataset("test_push_dataset_test", data_list)
        dataset.update(type="detection")
        dataset1 = cls.clt.create_dataset("test_push_dataset_test1", data_list)
        dataset1.update(type="detection")
        dataset2 = cls.clt.create_dataset("test_push_dataset_test2", data_list)
        dataset2.update(type="detection")
        cls.project.attach_dataset(dataset)

    def test_get_experiment_platform_url(self,):
        base_url = self.clt.connexion.host
        self.assertEqual("{}/experiment/{}".format(base_url, self.experiment.id),
                         self.experiment.get_resource_url_on_platform())

    def test_logging(self,):
        self.experiment.start_logging_chapter("chapter 1")
        self.experiment.start_logging_buffer(5)
        self.experiment.end_logging_buffer()
        self.experiment.send_experiment_logging("chapter 1", "chapter 1")

    def test_update_job_status(self,):
        self.experiment.update_job_status("success")

    def test_launch(self,):
        self.experiment.launch()

    def test_publish(self,):
        self.experiment.publish("test_model_1")
        model = self.clt.get_model("test_model_1")
        model.delete()

    def test_store_with_path_and_download(self,):
        self.experiment.store("small_file", "tests/files/test_dir", True)
        self.experiment.store("large_file", "tests/files/lg_test_file.pb")
        os.remove("tests/files/test_dir.zip")
        self.experiment.download("small_file", "tests/files/dl_test_dir")
        self.assertTrue(os.path.isfile("tests/files/dl_test_dir/test_dir.zip"))
        os.remove("tests/files/dl_test_dir/test_dir.zip")
        self.experiment.download("large_file", "tests/files/dl_test_dir")
        self.assertTrue(os.path.isfile(
            "tests/files/dl_test_dir/lg_test_file.pb"))
        os.remove("tests/files/dl_test_dir/lg_test_file.pb")
        self.experiment.store(
            "small_file", "tests/files/sm_test_file.config")  # test update file
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.experiment.download("not_a_file")
        files = self.experiment.list_artifacts()
        files = [fle["name"] for fle in files]
        self.assertIn("small_file", files)
        self.assertIn("large_file", files)
        self.project.get_experiment("test_experiment_1", with_artifacts=True)
        shutil.rmtree("test_experiment_1")
        self.project.get_experiment(
            "test_experiment_1", with_artifacts=True, tree=True)
        shutil.rmtree("test_experiment_1")
        self.experiment.delete_artifact("small_file")
        self.experiment.delete_all_artifacts()

    def test_log(self,):
        self.experiment.log("value", 25, 'value')  # test log single value
        self.experiment.log("image", 'tests/files/test.png',
                            'image')  # test log image
        self.experiment.log("line", [1, 2, 3, 4], "line")  # test log line
        # test log replace value
        self.experiment.log("value", 26, "value", True)
        # test log update no replace value
        self.experiment.log("value", 26, "value")
        self.experiment.log("line", [5, 6, 7], "line")  # test log append lines
        log_list = self.experiment.list_logs()
        log_list = [dt["name"] for dt in log_list]
        self.assertIn("value", log_list)
        self.assertIn("line", log_list)
        self.assertIn("image", log_list)
        self.experiment.delete_log("value")
        self.experiment.delete_all_logs()

    def test_download_annotations_pic_train_test_split(self,):
        dataset = self.clt.get_dataset("test_push_dataset_test", "first")
        dataset.upload_annotations_from_file(
            ann_path="tests/files/test_annotations_rectangle.json", ann_format="PICSELLIA")
        experiment = self.project.create_experiment(
            name="test_experiment_2", dataset=dataset)
        experiment.download_annotations()

        with self.assertRaises(exceptions.InvalidQueryError):
            experiment.download_annotations("random")

        experiment.png_dir = "tests/files/png_dir"
        experiment.download_pictures()
        experiment.generate_labelmap()
        experiment.train_test_split()
        experiment = self.project.get_experiment("test_experiment_1")
        experiment.dict_annotations = {}
        experiment.base_dir = ""
        with self.assertRaises(exceptions.ResourceNotFoundError):
            experiment.generate_labelmap()  # test no 'images' key in dict annotations
        with self.assertRaises(exceptions.ResourceNotFoundError):
            experiment.download_pictures()  # test no 'images' key in dict annotations
        with self.assertRaises(exceptions.ResourceNotFoundError):
            experiment.train_test_split()  # test no 'images' key in dict annotations
        shutil.rmtree("tests/files/png_dir")
        os.remove("label_map.pbtxt")

    def test_create_scan(self,):
        config = {
            'script': 'tests/__init__.py',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
            'data': ['tests/files/test.png'],
            'requirements': [
                {
                    'package': 'numpy',
                    'version': '0.0.1'
                }
            ]
        }
        self.project.create_scan("test_scan", config)

        config["script"] = "tests/files/not_a_file.txt"
        with self.assertRaises(FileNotFoundError):
            self.project.create_scan("test_scan_no_file", config)

    def test_create_scan_large_script(self,):
        config = {
            'script': 'tests/files/lg_test_file.pb',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
        }
        self.project.create_scan("test_scan_large", config)

    def test_create_scan_2(self,):
        config = {
            'image': 'picsellpn/simple-run:1.0',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
            'requirements': 'tests/files/test_req.txt'
        }
        self.project.create_scan("test_scan_2", config)

    def test_create_scan_error(self,):
        config = {
            'image': 'picsellpn/simple-run:1.0',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
            'requirements': {
                'package': 'numpy',
                'version': '0.0.1'
            }
        }
        with self.assertRaises(exceptions.InvalidQueryError):
            self.project.create_scan("test_scan_3", config)

    def test_create_scan_and_test_runs(self,):
        config = {
            'image': 'picsellpn/simple-run:1.0',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
        }
        scan = self.project.create_scan("test_scan_run_1", config)
        run = scan.get_next_run()
        experiment = run.get_experiment()
        self.assertEqual(experiment.name, "test_scan_run_1-0")
        run = scan.get_run_by_id(run.id)
        run.update(status="success")
        run.end()

    def test_run_download_data_script_and_req(self,):
        config = {
            'script': 'tests/__init__.py',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
            'data': ['tests/files/test.png'],
            'requirements': [
                {
                    'package': 'requests',
                    'version': '2.23.0'
                }
            ]
        }
        scan = self.project.create_scan("test_scan_run_2", config)
        run = scan.get_next_run()
        run.download_script()
        run.download_data()
        run.install_requirements()

    def test_generate_requirements(self,):
        config = {
            'script': 'tests/__init__.py',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
            'requirements': "tests/files/wrong_test_req.txt"
        }
        with self.assertRaises(exceptions.ResourceNotFoundError):
            # wrong requirements path
            self.project.create_scan("test_scan_run_3", config)
        config["requirements"] = "tests/files/bad_test_req.txt"
        config["requirements"] = "tests/files/test_req.txt"
        self.project.create_scan("test_scan_run_3", config)

    def test_create_scan_and_launch(self,):
        config = {
            'image': 'picsellpn/simple-run:1.0',
            'execution': {
                'type': 'agents'
            },
            'strategy': 'grid',
            'metric': {
                'name': 'Loss-total_loss',
                'goal': 'minimize'
            },
            'parameters': {
                'batch_size': {
                    'values': [2, 4, 8],
                },
            },
        }
        scan = self.project.create_scan("test_scan_run_4", config)
        with self.assertRaises(exceptions.InvalidQueryError):
            scan.launch()


class TestClearAssets(TestInitializedClientSDK):

    def test_clear_datasets(self):
        try:
            datasets = self.clt.list_datasets()
        except exceptions.ResourceNotFoundError:
            datasets = []
        for dataset in datasets:
            dataset.delete()

    def test_clear_data(self):
        try:
            datas = self.clt.get_datalake().list_data()
        except exceptions.NoDataError:
            datas = []

        for data in datas:
            data.delete()


    def test_clear_models(self):
        try:
            models = self.clt.list_models()
        except Exception:
            models = []

        for model in models:
            model.delete()

    def test_clear_projects(self):
        try:
            projects = self.clt.list_projects()
        except Exception:
            projects = []

        for project in projects:
            project.delete()

    def test_clear_deployments(self):
        try:
            deployments = self.clt.list_deployments()
        except Exception:
            deployments = []

        for deployment in deployments:
            deployment.delete()



class TestDeployment(TestInitializedClientSDK):

    @classmethod
    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def setUpClass(cls) -> None:
        super().setUpClass()
        datalake = cls.clt.get_datalake()
        models = cls.clt.list_models()
        for model in models:
            model.delete()
        try:
            datalist = datalake.list_data()
        except exceptions.NoDataError:
            datalist = []
        for one_data in datalist:
            one_data.delete()
        try:
            deployments = cls.clt.list_deployments()
        except exceptions.ResourceNotFoundError:
            deployments = []
        for deployment in deployments:
            deployment.delete()

        project_list = cls.clt.list_projects()
        for project in project_list:
            project.delete()

        files = os.listdir('tests/files/test_image_dir')
        cls.files = ['tests/files/test_image_dir/' + fle for fle in files]
        cls.datalake = cls.clt.get_datalake()
        cls.datalake.upload_data(cls.files, ["tag3"])
        data_list = datalake.fetch_data(quantity=2)
        cls.dataset = cls.clt.create_dataset("datatest", data_list)
        cls.dataset.update('detection')
        cls.dataset2 = cls.clt.create_dataset("datatest2", data_list)
        cls.dataset2.update('detection')
        cls.dataset_no_type = cls.clt.create_dataset("dataset_no_type", data_list)
        cls.dataset_wrong_type = cls.clt.create_dataset("dataset_wrong_type", data_list)
        cls.dataset_wrong_type.update('detectionKOK')
        cls.project = cls.clt.create_project("test_project_1")
        cls.experiment = cls.project.create_experiment(
            name="test_experiment_1")
        cls.parameters = {
            "param1": 1
        }
        cls.experiment.log('parameters', data=cls.parameters, type="table")
        cls.model = cls.clt.create_model("test_model", "detection")
        cls.model.store('model-latest', 'tests/files/test_dir', True)
        cls.model_with_sources = cls.clt.create_model("test_model_with_source", "detection")
        cls.model_with_sources.update(
            source_dataset=cls.dataset,
        )
        cls.model_with_sources.update(
            source_experiment=cls.experiment
        )
        cls.model_with_sources.store('model-latest', 'tests/files/test_dir', True)
        cls.deployment = cls.model_with_sources.deploy(
            config={
                "min_det_threshold": 0.2
            }
        )
        cls.deployment2 = cls.model.deploy(
            config={
                "min_det_threshold": 0.2
            }
        )


    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def test_get_deployment_by_name(self):
        fetched_deployment = self.clt.get_deployment(
            name=self.deployment.name
        )
        self.assertEqual(self.deployment.id, fetched_deployment.id)
        self.assertEqual(self.deployment.name, fetched_deployment.name)

    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def test_get_deployment_by_id(self):
        fetched_deployment = self.clt.get_deployment_by_id(
            id=str(self.deployment.id)
        )
        self.assertEqual(self.deployment.id, fetched_deployment.id)
        self.assertEqual(self.deployment.name, fetched_deployment.name)

    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def test_get_deployment_by_name_not_exists(self):
        with self.assertRaises(exceptions.ResourceNotFoundError): 
            name = "stupid-name"
            fetched_deployment = self.clt.get_deployment(
                name=name
            )

    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def test_get_deployment_by_id_not_exists(self):
        with self.assertRaises(exceptions.ResourceNotFoundError): 
            stupid_id = str(uuid.uuid4())
            fetched_deployment = self.clt.get_deployment_by_id(
                id=stupid_id
            )

    @patch('picsellia.sdk.deployment.JwtServiceConnexion', Mock(spec=JwtServiceConnexion))
    def test_list_deployments(self):
        deployments = self.clt.list_deployments()
        self.assertEqual(len(deployments), 2)
        self.assertIn(self.deployment.id, [d.id for d in deployments])
        self.assertIn(self.deployment.name, [d.name for d in deployments])
        self.assertIn(self.deployment2.id, [d.id for d in deployments])
        self.assertIn(self.deployment2.name, [d.name for d in deployments])

    def test_init_feedback_loop_wrong_dataset_type(self):
        with self.assertRaises(exceptions.InvalidQueryError):
            self.deployment.init_feedback_loop(
                dataset=self.dataset_wrong_type
            )

    def test_init_feedback_loop_no_dataset_type(self):
        with self.assertRaises(exceptions.InvalidQueryError):
            self.deployment.init_feedback_loop(
                dataset=self.dataset_no_type
            )

    def test_init_feedback_loop(self):
        self.deployment.init_feedback_loop(
            dataset=self.dataset
        )

    def test_update_feedback_loop(self):
        self.deployment2.init_feedback_loop(
            dataset=self.dataset
        )
        self.deployment2.update_feedback_loop(
            dataset=self.dataset2
        )

    def test_init_continous_training(self):
        self.deployment.init_continuous_training(
            project=self.project,
            dataset=self.dataset,
            model=self.model_with_sources,
            parameters=self.parameters,
            training_type="experiment",
            feedback_loop_trigger=100
        )

    def test_activate_continuous_training(self):
        self.deployment.activate_continuous_training()

    def test_detest_activate_continuous_training(self):
        self.deployment.deactivate_continuous_training()

    def test_activate_continuous_training_not_set(self):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.deployment2.activate_continuous_training()

    def test_deactivate_continuous_training_not_set(self):
        with self.assertRaises(exceptions.ResourceNotFoundError):
            self.deployment2.deactivate_continuous_training()
    
    def test_retrieve_information_mocked_oracle(self):
        
        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : {"pk": "iddeployment", "name": "name-test"}}
        response.status_code = 200
        self.deployment.oracle_connexion.get.return_value = response

        self.assertEqual({"deployment" : {"pk": "iddeployment", "name": "name-test"}}, self.deployment.retrieve_information())

        self.deployment.oracle_connexion.get.assert_called_once_with(path='/deployment/{}'.format(self.deployment.id))

        self.deployment.oracle_connexion = old_connexion

    def test_retrieve_information_when_no_connexion(self):
        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = None
        
        with self.assertRaises(AssertionError):
            self.deployment.retrieve_information()

        self.deployment.oracle_connexion = old_connexion
    
    def test_predict(self):
                
        old_connexion = self.deployment.serving_connexion
        self.deployment.serving_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : {"pk": "iddeployment", "predicted": True}}
        response.status_code = 200
        self.deployment.serving_connexion.post.return_value = response
        
        self.deployment.predict("tests/files/test.png")

        self.deployment.serving_connexion.post.assert_called_once_with(path='/predict/{}'.format(self.deployment.id), files=ANY)

        self.deployment.serving_connexion = old_connexion

    def test_monitor(self):
                
        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : {"pk": "iddeployment", "predicted": True}}
        response.status_code = 200
        self.deployment.oracle_connexion.post.return_value = response
        
        self.deployment.monitor(self.model, "tests/files/test.png", 0.8, "test-picture-id", "test.png", 100, 100, "test-sdk", {"test": 1}, ClassificationPredictionFormat(1, 50))

        self.deployment.oracle_connexion.post.assert_called_once_with(path='/deployment/{}/add'.format(self.deployment.id), data=ANY)

        self.deployment.oracle_connexion = old_connexion

    def test_monitor_with_shadow(self):
                
        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : {"pk": "iddeployment", "predicted": True}}
        response.status_code = 200
        self.deployment.oracle_connexion.post.return_value = response
        
        self.deployment.monitor(self.model, "tests/files/test.png", 0.8, "test-picture-id", "test.png", 100, 100, "test-sdk", {"test": 1}, ClassificationPredictionFormat(1, 50), shadow_model=self.model_with_sources, shadow_latency=0.3, shadow_raw_predictions=ClassificationPredictionFormat(1, 78))

        self.deployment.oracle_connexion.post.assert_called_once_with(path='/deployment/{}/add'.format(self.deployment.id), data=ANY)

        self.deployment.oracle_connexion = old_connexion

    def test_monitor_with_shadow_but_not_shadow_latency(self):
                
        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : {"pk": "iddeployment", "predicted": True}}
        response.status_code = 200
        self.deployment.oracle_connexion.post.return_value = response
        
        with self.assertRaises(exceptions.ImpossibleAction):
            self.deployment.monitor(self.model, "tests/files/test.png", 0.8, "test-picture-id", "test.png", 100, 100, "test-sdk", {"test": 1}, ClassificationPredictionFormat(1, 50), shadow_model=self.model_with_sources, shadow_raw_predictions=ClassificationPredictionFormat(1, 78))

        self.assertEqual(self.deployment.oracle_connexion.post.called, 0)

        self.deployment.oracle_connexion = old_connexion


    def test_get_stats_on_predictions(self):

        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : { "data" : {"val": "ue"}, "deployment_id" : self.deployment.id}}
        response.status_code = 200
        self.deployment.oracle_connexion.get.return_value = response

        self.deployment.get_stats(ServiceMetrics.PREDICTIONS_DATA)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/predictions/stats'.format(self.deployment.id), data={"service": "metrics"})

        self.deployment.get_stats(ServiceMetrics.PREDICTIONS_OUTLYING_SCORE)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/predictions/stats'.format(self.deployment.id), data={"service": "ae_outlier"})
        
        self.deployment.get_stats(ServiceMetrics.REVIEWS_OBJECT_DETECTION_STATS)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/predictions/stats'.format(self.deployment.id), data={"service": "object_detection"})

        self.deployment.get_stats(ServiceMetrics.REVIEWS_CLASSIFICATION_STATS)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/predictions/stats'.format(self.deployment.id), data={"service": "classification"})

        self.deployment.get_stats(ServiceMetrics.REVIEWS_LABEL_DISTRIBUTION_STATS)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/predictions/stats'.format(self.deployment.id), data={"service": "label_distribution"})

        self.deployment.oracle_connexion = old_connexion

       
    def test_get_stats_on_deployment(self):

        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : { "data" : {"val": "ue"}, "deployment_id" : self.deployment.id}}
        response.status_code = 200
        self.deployment.oracle_connexion.get.return_value = response

        self.deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={"service": "label_distribution"})

        self.deployment.get_stats(ServiceMetrics.AGGREGATED_OBJECT_DETECTION_STATS)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={"service": "object_detection"})
        
        self.deployment.get_stats(ServiceMetrics.AGGREGATED_PREDICTIONS_DATA)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={"service": "metrics"})

        self.deployment.get_stats(ServiceMetrics.AGGREGATED_DRIFTING_PREDICTIONS)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={"service": "ks_drift"})

        self.deployment.oracle_connexion = old_connexion

    def test_get_stats_on_deployment_with_warning(self):

        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : { "data" : {"val": "ue"}, "deployment_id" : self.deployment.id}, "infos": {"info": "outdated"}}
        response.status_code = 200
        self.deployment.oracle_connexion.get.return_value = response

        self.deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION)
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={"service": "label_distribution"})

        self.deployment.oracle_connexion = old_connexion

    def test_get_stats_on_deployment_with_all_filter(self):

        old_connexion = self.deployment.oracle_connexion
        self.deployment.oracle_connexion = Mock(spec=JwtServiceConnexion)

        response = Mock(spec=Response)
        response.json.return_value = {"deployment" : { "data" : {"val": "ue"}, "deployment_id" : self.deployment.id}}
        response.status_code = 200
        self.deployment.oracle_connexion.get.return_value = response

        self.deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION, model=self.model, from_timestamp=1, to_timestamp=2, since=10, includes=[], excludes=[], tags='tag:value')
        self.deployment.oracle_connexion.get.assert_called_with(path='/deployment/{}/stats'.format(self.deployment.id), data={'service': 'label_distribution', 'model_id': self.model.id, 'from_timestamp': 1, 'to_timestamp': 2, 'since': 10, 'includes': [], 'excludes': [], 'tags': 'tag:value'})

        self.deployment.oracle_connexion = old_connexion




from functools import partial
import json
import logging
import os
from typing import Any, List, Union
from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia import exceptions
from picsellia import utils
import threading
from picsellia import pxl_multithreading as mlt
from pathlib import Path
import requests
from picsellia.sdk.model import Model
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Experiment(Dao):

    def __init__(self, connexion: Connexion, project_token: str, id: str, name: str, files: dict) -> None:
        super().__init__(connexion)
        self.project_token = project_token
        self.id = id
        self.name = name
        self.files = files

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/experiment/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/experiment/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def update(self, **kwargs) -> None:
        """Update this experiment with given keys.

        Examples:
            ```python
                my_experiment.update(description="First try Yolov5")
            ```
        """
        assert kwargs != {}, "You shall give some keys to update"
        data = json.dumps(kwargs)
        self.connexion.patch('/sdk/v1/experiment/{}'.format(self.id), data=data)

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete this experiment

        Examples:
            ```python
                my_experiment.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/experiment/{}'.format(self.id))

    @exception_handler
    @beartype
    def list_artifacts(self,) -> List[dict]:
        """List all stored artifacts in the experiment.

        Examples:
            ```python
                artifacts = my_experiment.list_artifacts()
            ```

        Returns:
            A list of artifact infos as dict like
            {
                'id':str,
                'name':str,
                'object_name':str,
                'large':bool
            }
        """
        r = self.connexion.get(
            '/sdk/v1/experiment/{}/artifacts'.format(self.id)).json()
        return r["files"]

    @exception_handler
    @beartype
    def delete_all_artifacts(self,) -> None:
        """Delete all stored artifacts for experiment

        :warning: **DANGER ZONE**: This will definitely remove the artifacts from our servers

        Examples:
            ```python
                experiment.delete_all_artifacts()
            ```
        """
        self.connexion.delete('/sdk/v1/experiment/{}/artifacts'.format(self.id))

    @exception_handler
    @beartype
    def _create_file(self, name: str = "", object_name: str = "", large: bool = False) -> dict:
        """Creates artifact and attach it to experiment on Platform.

        Examples:
            ```python
                self._create_file()
            ```
        Arguments:
            name (str, optional): filename. Defaults to "".
            object_name (str, optional): s3 object name. Defaults to "".
            large (bool, optional): >5Mb or not. Defaults to False.

        Returns:
            An artifact info as dict like
            {
                "id":str,
                "name":str,
                "object_name":str,
                "large":bool
            }
        """
        data = json.dumps({'0': {
            'name': name,
            'object_name': object_name,
            'large': large
        }
        })
        r = self.connexion.post(
            '/sdk/v1/experiment/{}/artifact'.format(self.id), data=data)
        return r.json()

    @exception_handler
    @beartype
    def _create_or_update_file(self, file_name: str = "", path: str = "", **kwargs) -> None:
        """Decides whether to create or update an artifact

        Examples:
            ```python
                self._create_or_update_file()
            ```
        Arguments:
            file_name (str, optional): filename. Defaults to "".
            path (str, optional): full path to file. Defaults to "".
        """
        try:
            stored = self.get_artifact(file_name)
        except exceptions.ResourceNotFoundError:
            stored = []
        if stored == []:
            self._create_file(
                file_name, kwargs["object_name"], kwargs["large"])
        else:
            self._update_artifact(name=file_name, **kwargs)

    @exception_handler
    @beartype
    def store(self, name: str, path: str = None, zip: bool = False) -> None:
        """Store an artifact and attach it to the experiment.

        Examples:
            ```python
                # Zip and store a folder as an artifact for the experiment
                # you can choose an arbitrary name or refer to our 'namespace'
                # for certain artifacts to have a custom behavior

                trained_model_path = "my_experiment/saved_model"
                experiment.store("model-latest", trained_model_path, zip=True)
            ```
        Arguments:
            name (str): name for the artifact. Defaults to "".
            path (str): path to the file or folder. Defaults to None.
            zip (bool, optional): Whether or not to compress the file to a zip file. Defaults to False.

        Raises:
            FileNotFoundError: No file found at the given path
        """
        if path is not None:
            if zip:
                path = utils.zipdir(path)
            filesize = Path(path).stat().st_size
            if filesize < 5*1024*1024:
                filename = os.path.split(path)[-1]
                if name == 'model-latest':  # pragma: no cover
                    object_name = os.path.join(self.id, '0', filename)
                else:
                    object_name = os.path.join(self.id, filename)
                resp = self.connexion._send_file(path, object_name)
                if resp.status_code == 204:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=False)
            else:
                filename = os.path.split(path)[-1]
                if name == 'model-latest':  # pragma: no cover
                    object_name = os.path.join(self.id, '0', filename)
                else:
                    object_name = os.path.join(self.id, filename)
                resp = self.connexion._send_large_file(path, object_name)
                if resp.status_code == 201:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=True)
        else:  # pragma: no cover
            if name == 'config':
                if not os.path.isfile(os.path.join(self.config_dir, "pipeline.config")):
                    raise FileNotFoundError("No config file found")
                path = os.path.join(self.config_dir, "pipeline.config")
                object_name = os.path.join(self.id, "pipeline.config")
                resp = self.connexion._send_file(path, object_name)
                if resp.status_code == 204:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=False)
            elif name == 'checkpoint-data-latest':
                file_list = os.listdir(self.checkpoint_dir)
                ckpt_id = max([int(p.split('-')[1].split('.')[0])
                              for p in file_list if 'index' in p])
                ckpt_data_file = None
                for f in file_list:
                    if "{}.data".format(ckpt_id) in f:
                        ckpt_data_file = f
                if ckpt_data_file is None:
                    raise exceptions.ResourceNotFoundError(
                        "Could not find matching data file with index")
                path = os.path.join(self.checkpoint_dir, ckpt_data_file)
                object_name = os.path.join(self.id, ckpt_data_file)
                resp = self.connexion._send_large_file(path, object_name)
                if resp.status_code == 201:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=True)
            elif name == 'checkpoint-index-latest':
                file_list = os.listdir(self.checkpoint_dir)
                ckpt_id = max([int(p.split('-')[1].split('.')[0])
                              for p in file_list if 'index' in p])
                ckpt_index = "ckpt-{}.index".format(ckpt_id)
                path = os.path.join(self.checkpoint_dir, ckpt_index)
                object_name = os.path.join(self.id, ckpt_index)
                resp = self.connexion._send_file(path, object_name)
                if resp.status_code == 204:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=False)
            elif name == 'model-latest':  # pragma: no cover
                file_path = os.path.join(
                    self.exported_model_dir, 'saved_model')
                path = utils.zipdir(file_path)
                object_name = os.path.join(self.id, '0', 'saved_model.zip')
                resp = self.connexion._send_large_file(path, object_name)
                if resp.status_code == 201:
                    self._create_or_update_file(
                        name, path, object_name=object_name, large=True)

    @exception_handler
    @beartype
    def download(self, name: str, target_path: str = '', large: bool = None) -> None:
        """Download an experiment's artifact to a given target_path.

        Examples:
            ```python
                experiment.download("model-latest", "myDir")
                file_list = os.path.listdir("myDir")
                print(file_list)
                >>> ["saved_model.zip"]
            ```
        Arguments:
            name (str): Name of the artifact to download
            target_path (str, optional): Path to download the file to, default to cwd. Defaults to ''.
            large (bool, optional): If download fail with large=False, try again with large=True.
                Defaults to None.
        """
        # Will raise NotFoundError if name is not found
        f = self.get_artifact(name)

        object_name = f["object_name"]
        if large is None:
            large = f["large"]
        filename = os.path.split(object_name)[-1]
        if self.connexion.download_some_file(large, object_name, os.path.join(target_path, filename)):
            logger.info('{} downloaded successfully'.format(filename))
        else:
            logger.error("Could not download {} file".format(filename))

    @exception_handler
    @beartype
    def get_artifact(self, name: str) -> dict:
        """Retrieve an artifact information.

        Examples:
            ```python
                model_artifact = experiment.get_artifact("model-latest")
                print(model_artifact)
                >>> {
                        "id":137,
                        "name":"model-latest",
                        "object_name":"d67924a0-7757-48ed-bf7a-322b745e917e/saved_model.zip",
                        "large":True
                }
            ```
        Arguments:
            name (str): Name of the artifact to retrieve

        Returns:
            List of artifact info as dict like
            {
                "id":str,
                "name":str,
                "object_name":str,
                "large":bool
            }
        """
        r = self.connexion.get(
            '/sdk/v1/experiment/{}/artifact?name={}'.format(self.id, name)).json()
        return r["file"]

    @exception_handler
    @beartype
    def _update_artifact(self, name: str, **kwargs) -> None:
        """Update artifact property

        Examples:
            ```python
                self._update_artifact()
            ```
        Arguments:
            name (str): name of the artifact to update
        """
        data = json.dumps(kwargs)
        self.connexion.patch('/sdk/v1/experiment/{}/artifact?name={}'.format(self.id, name), data=data).json()

    @exception_handler
    @beartype
    def delete_artifact(self, name: str) -> None:
        """Delete artifact from experiment

        :warning: The file of the artifact will be removed from our servers.

        Examples:
            ```python
                experiment.delete_artifact("saved-model")
            ```
        Arguments:
            name (str): Name of the artifact
        """
        self.connexion.delete('/sdk/v1/experiment/{}/artifact?name={}'.format(self.id, name))

    @exception_handler
    @beartype
    def list_logs(self,) -> list:
        """List everything that has been logged.

        List everything that has been logged to an experiment using the .log() method.

        Examples:
            ```python
                logs = experiment.list_logs()
                print(logs[0])
                >>> {
                    "id":137,
                    "date_created":"2021-09-06T10:47:56.569221Z",
                    "last_update":"2021-09-06T10:47:56.569221Z",
                    "name":"parameters",
                    "type":"table",
                    "data":{
                        "batch_size":4,
                        "epochs":1000
                    }
                }
            ```

        Returns:
            A list of logs objects as dict like {
                "id":int,
                "date_created":str,
                "last_update":str,
                "name":str,
                "type":str,
                "data":Union[list, dict]
            }
        """
        r = self.connexion.get('/sdk/v1/experiment/{}/logs'.format(self.id)).json()
        return r["logs"]

    @exception_handler
    @beartype
    def delete_all_logs(self,) -> None:
        """Delete everything that has been logged.

        Delete everything that has been logged (using .log()) into this experiment  method.

        Examples:
            ```python
                experiment.delete_all_logs()
            ```
        """
        self.connexion.delete('/sdk/v1/experiment/{}/logs'.format(self.id))

    @exception_handler
    @beartype
    def create_log(self, name: str = "", data: Any = {}, type: str = None) -> None:
        """Create log in a experiment.

        Arguments:
            name (str, optional): Name of data. Defaults to "".
            data (dict, optional): Data content. Defaults to {}.
            type (str, optional): Type of data. Defaults to None.
        """
        data = json.dumps({'0': {
            'name': name,
            'data': data,
            'type': type
        }
        })
        self.connexion.put(
            '/sdk/v1/experiment/{}/log'.format(self.id), data=data).json()

    @exception_handler
    @beartype
    def get_log(self, name: str) -> Union[dict, list]:
        """Get data for a given log in this experiment

        Examples:
            ```python
                parameters = experiment.get_log("parameters")
                print(parameters)
                >>> {
                    "batch_size":4,
                    "epochs":1000
                }
            ```
        Arguments:
            name (str): name of the log to retrieve

        Returns:
            A dict or list of the log you logged with given name
        """
        r = self.connexion.get(
            '/sdk/v1/experiment/{}/log?name={}'.format(self.id, name)).json()
        data_name = r["log"]["name"]
        if not hasattr(self, "logged_log_names"):
            self.logged_log_names = []
        if data_name not in self.logged_log_names:
            self.logged_log_names.append(data_name)

        return r["log"]["data"]

    @exception_handler
    @beartype
    def append_log(self, name: str, **kwargs) -> None:
        """Appends value to log with given name.

        Arguments:
            name (str): name of the log is mandatory

        """
        data = json.dumps(kwargs)
        self.connexion.post(
            '/sdk/v1/experiment/{}/log?name={}'.format(self.id, name), data=data).json()

    @exception_handler
    @beartype
    def update_log(self, name: str, **kwargs):
        """Update log with given name.

        Arguments:
            name (str): name of the log.

        Returns:
            An updated log object as dict like {
                "id":int,
                "date_created":str,
                "last_update":str,
                "name":str,
                "type":str,
                "data":Union[list, dict]
            }
        """
        data = json.dumps(kwargs)
        r = self.connexion.patch(
            '/sdk/v1/experiment/{}/log?name={}'.format(self.id, name), data=data).json()
        return r

    @exception_handler
    @beartype
    def delete_log(self, name: str) -> None:
        """Delete log with given name in the experiment

        Examples:
            ```python
                experiment.delete_log("parameters")
            ```
        Arguments:
            name (str): name of the log to delete
        """
        self.connexion.delete(
            '/sdk/v1/experiment/{}/log?name={}'.format(self.id, name))

    @exception_handler
    @beartype
    def log(self, name: str, data: Any, type: str = None, replace: bool = False) -> None:
        """Log (record) anything to an experiment.

        Record something to an experiment.
        It will then be saved and displayed.

        Examples:
            ```python
                parameters = {
                    "batch_size":4,
                    "epochs":1000
                }
                exp.log("parameters", parameters, type="table")
            ```
        Arguments:
            name (str): Name of the log.
            data (Any): Data to be saved.
            type (str, optional): Type of the data to log.
                                  This will condition how it is displayed in the experiment dashboard. Defaults to None.
            replace (bool, optional): Whether to replace the current value of the log. Defaults to False.

        Raises:
            Exception: Impossible to upload the file when logging an image.
        """
        try:
            if not hasattr(self, "logged_log_names") or name not in self.logged_log_names:
                stored = self.get_log(name)
            else:
                stored = self.logged_log_names
        except exceptions.ResourceNotFoundError:
            stored = []
        if type == 'value':
            data = {'value': data}
        if type == 'image':
            object_name = os.path.join(self.id, data)
            response = self.connexion.get_presigned_url(
                method='post', object_name=object_name, bucket_model=True)
            with open(data, 'rb') as f:
                files = {'file': (data, f)}
                http_response = requests.post(
                    response['url'], data=response['fields'], files=files)
                if http_response.status_code == 204:
                    data = {'object_name': object_name}
                else:  # pragma: no cover
                    raise Exception(
                        "Impossible to log image, can't upload file, please contact us.")
        if stored == []:
            assert type is not None, \
                "Please specify a type for your data vizualization, check the docs to see all available types"
            self.create_log(name, data=data, type=type)
        elif stored != [] and replace:
            self.update_log(name, data=data, type=type)
        elif stored is not [] and not replace and type == 'line':
            threading.Thread(target=self.append_log,
                             kwargs={'name': name, 'data': data, 'type': type}).start()
        elif stored != [] and not replace and type != 'line':
            self.update_log(name, data=data, type=type)

    @exception_handler
    @beartype
    def send_experiment_logging(self, log: Union[str, list], part: str, final: bool = False, special: Union[str, bool, list] = False) -> None:
        """Send a logging experiment to the experiment .

        Arguments:
            log (str): [Log content]
            part (str): [Logging Part]
            final (bool, optional): [True if Final line]. Defaults to False.
            special (bool, optional): [True if special log]. Defaults to False.

        Raises:
            exceptions.NetworkError: [Picsellia Platform not responding]
        """
        if not hasattr(self, 'line_nb'):
            self.line_nb = 0
        to_send = {
            "experiment_id": self.id,
            "line_nb": self.line_nb,
            "log": log,
            "final": final,
            "part": part,
            "special": special
        }
        self.line_nb += 1
        try:
            self.connexion.post(
                '/sdk/v1/experiment/{}/logging'.format(self.id), data=json.dumps(to_send))
        except Exception:  # pragma: no cover
            logger.error("Unable to send logs to platform. Log : {}".format(log))

    @exception_handler
    @beartype
    def start_logging_chapter(self, name: str) -> None:
        """Print a log entry to the log .

        Arguments:
            name (str): Chapter name
        """
        print('--#--' + name)

    @exception_handler
    @beartype
    def start_logging_buffer(self, length: int = 1) -> None:
        """Start logging buffer .

        Arguments:
            length (int, optional): Buffer length. Defaults to 1.
        """
        print('--{}--'.format(str(length)))
        self.buffer_length = length

    @exception_handler
    @beartype
    def end_logging_buffer(self,) -> None:
        """End the logging buffer .
        """
        print('---{}---'.format(str(self.buffer_length)))

    @exception_handler
    @beartype
    def update_job_status(self, status: str) -> None:
        """Update the job status.

        Arguments:
            status (str): [Status to send]

        Raises:
            exceptions.NetworkError: [Picsellia Platform not responding]
        """
        to_send = {
            "status": status,
        }
        self.connexion.patch(
            '/sdk/v1/experiment/{}/job_status'.format(self.id), data=json.dumps(to_send))

    @exception_handler
    @beartype
    def _convert_response_to_model(self, response: dict) -> Model:
        return Model(
            self.connexion,
            response["model_id"],
            response["network_name"],
            response["type"],
        )

    @exception_handler
    @beartype
    def publish(self, name: str) -> Model:
        """Publish an Experiment as a Model to your registry.

        Examples:
            ```python
                model = experiment.publish("awesome-model")
                model.update(framework="tensorflow")
            ```
        Arguments:
            name (str): Target Name for the model in the registry.

        Returns:
            A (Model) just created from the experiment
        """
        data = json.dumps({
            "name": name
        })
        r = self.connexion.post(
            '/sdk/v1/experiment/{}/publish'.format(self.id), data=data).json()
        model = self._convert_response_to_model(r["network"])
        logger.info("Experiment published as a model with name {}".format(name))
        return model

    @exception_handler
    @beartype
    def launch(self, gpus: int = 0) -> None:
        """Launch a job on a remote environment with this experiment.

        :information-source: The remote environment has to be setup prior launching the experiment.
        It defaults to our remote training engine.

        Examples:
            ```python
                experiment.launch()
            ```
        Arguments:
            gpus (int, optional): Number of GPU to use for the training. Defaults to 0.
        """
        data = json.dumps({
            "gpus": gpus,
        })
        self.connexion.post('/sdk/v1/experiment/{}/launch'.format(self.id), data=data)
        logger.info("Job launched successfully")

    def _setup_dirs(self):
        """Create the directories for the project.
        """
        self.base_dir = self.name
        self.metrics_dir = os.path.join(self.base_dir, 'metrics')
        self.png_dir = os.path.join(self.base_dir, 'images')
        self.checkpoint_dir = os.path.join(self.base_dir, 'checkpoint')
        self.record_dir = os.path.join(self.base_dir, 'records')
        self.config_dir = os.path.join(self.base_dir, 'config')
        self.results_dir = os.path.join(self.base_dir, 'results')
        self.exported_model_dir = os.path.join(self.base_dir, 'exported_model')

        if not os.path.isdir(self.name):
            logger.debug("No directory for this project has been found, creating directory and sub-directories...")
            os.mkdir(self.name)

        self._create_dir(self.base_dir)
        self._create_dir(self.png_dir)
        self._create_dir(self.checkpoint_dir)
        self._create_dir(self.metrics_dir)
        self._create_dir(self.record_dir)
        self._create_dir(self.config_dir)
        self._create_dir(self.results_dir)
        self._create_dir(self.exported_model_dir)

    @exception_handler
    @beartype
    def _create_dir(self, dir_name: str) -> None:
        """Create a directory if it doesn t exist.

        Arguments:
            dir_name (str): [directory name]
        """
        if not os.path.isdir(dir_name):
            os.mkdir(dir_name)

    @exception_handler
    @beartype
    def download_annotations(self, option: str = "all"):
        """Download the annotation of the dataset attached to the experiment.
            The method stores the dict containing the annotations in experiment.dict_annotations attribute.

        Examples:
            ```python
                annotation_dict = experiment.download_annotations("accepted")
            ```
        Arguments:
            option (str, optional): Whether to download only accepted annotations or every annotations
                . Defaults to "all".

        Raises:
            exceptions.ResourceNotFoundError: No annotations for the attached dataset.

        Returns:
            A dict containing all the annotations in the Picsellia format.
        """
        logger.info("Downloading annotations ...")
        r = self.connexion.get(
            '/sdk/v1/experiment/{}/dl_annotations?type={}'.format(self.id, option))
        self.dict_annotations = r.json()
        if self.dict_annotations == None or len(self.dict_annotations.keys()) == 0 or "error" in self.dict_annotations:
            raise exceptions.ResourceNotFoundError("You don't have any annotations")

    @exception_handler
    @beartype
    def download_pictures(self, nb_threads: int = 20) -> None:
        """Download the picture from the dataset attached to the experiment.
            It download only the pictures corresponding to the annotations fetched with the .download_annotations
            method. This prevent to download unnecessary images from the dataset.
            If you wish to to download the whole dataset, we recommend you to use the
            client methods get_dataset() and .download() on the retrieved object.

        Examples:
            ```python
                experiment.download_pictures()
            ```

        Raises:
            exceptions.ResourceNotFoundError: You have to run the .download_annotations() method first.
        """
        if self.dict_annotations == None or "images" not in self.dict_annotations.keys():
            raise exceptions.ResourceNotFoundError(
                "Please run download_annotations function first")

        logger.info("Downloading images ...")

        if not os.path.isdir(self.png_dir):
            os.makedirs(self.png_dir)

        print("-----")
        images = self.dict_annotations["images"]

        f = partial(mlt.mlt_download_list_external_picture, self.png_dir)
        mlt.do_multiprocess_things(f, images, nb_threads)

        print("--*--")
        logger.info("Images downloaded")

    @exception_handler
    @beartype
    def generate_labelmap(self,) -> None:
        """Generate a labelmap in the pbtxt format.

        Generate labelmap in the current working directory and save the dictionnary in
            experiment.label_map

        Examples:
            ```python
                experiment.generate_labelmap()
            ```

        Raises:
            exceptions.ResourceNotFoundError: You have to call the .download_annotations() first.
            exceptions.ResourceNotFoundError: Path to write labelmap file not found.
        """

        logger.debug("Generating labelmap ...")
        if not hasattr(self, 'base_dir'):
            self.base_dir = ""
        self.label_path = os.path.join(self.base_dir, "label_map.pbtxt")

        if "categories" not in self.dict_annotations:
            raise exceptions.ResourceNotFoundError(
                "Please run download_annotations() first")

        categories = self.dict_annotations["categories"]
        labels = {}
        try:
            with open(self.label_path, "w+") as labelmap_file:
                for k, category in enumerate(categories):
                    name = category["name"]
                    labelmap_file.write(
                        "item {\n\tname: \"" + name + "\"" + "\n\tid: " + str(k + 1) + "\n}\n")
                    labels[str(k + 1)] = name
                labelmap_file.close()
            logger.debug("Label_map.pbtxt created @ {}".format(self.label_path))

        except Exception:  # pragma: no cover
            raise exceptions.ResourceNotFoundError(
                "No directory found, please call checkout_network() or create_network() function first")

        self.label_map = labels

    @exception_handler
    @beartype
    def train_test_split(self, prop: float = 0.8) -> tuple:
        """Train test split

        Examples:
            ```python
                experiment.train_test_split()
            ```
        Arguments:
            prop (float, optional): Percentage of data for training set. Defaults to 0.8.

        Raises:
            exceptions.ResourceNotFoundError: No annotations. You have to run the .download_annotations() first.

        Returns:
            A tuple with all of this information (
                list of train picture files,
                list of test picture files,
                list of train picture ids,
                list of test picture ids,
                dict of repartition of classes for train set,
                dict of repartition of classes for test set,
                list of labels
            )
        """
        if "images" not in self.dict_annotations:
            raise exceptions.ResourceNotFoundError(
                "Please download annotations first")

        train_list = []
        eval_list = []
        train_list_id = []
        eval_list_id = []
        index_url = utils.train_valid_split_obj_simple(self.dict_annotations, prop)

        for info, idx in zip(self.dict_annotations["images"], index_url):
            pic_name = os.path.join(self.png_dir, info['external_picture_url'])
            if idx == 1:
                train_list.append(pic_name)
                train_list_id.append(info["internal_picture_id"])
            else:
                eval_list.append(pic_name)
                eval_list_id.append(info["internal_picture_id"])

        logger.debug("{} images used for training and {} images used for validation"
                     .format(len(train_list_id), len(eval_list_id)))

        label_train, label_test, cate = utils.get_labels_repartition_obj_detection(self.dict_annotations, index_url)
        return train_list, eval_list, train_list_id, eval_list_id, label_train, label_test, cate


    @exception_handler
    @beartype
    def _download_artifacts_and_prepare_tree(self, tree : bool, with_artifacts : bool):
        if tree:
            self._setup_dirs()

        if with_artifacts:
            if tree:
                self._download_artifacts_with_tree_for_experiment()
            else:
                self._download_artifacts_without_tree_for_experiment()

    @exception_handler
    @beartype
    def _download_artifacts_with_tree_for_experiment(self):
        for f in self.files:
            object_name = f["object_name"]
            name = f["name"]
            filename = os.path.split(f["object_name"])[-1]
            if name == 'checkpoint-data-latest':  # pragma: no cover
                path = os.path.join(self.checkpoint_dir, filename)
            elif name == 'checkpoint-index-latest':  # pragma: no cover
                path = os.path.join(self.checkpoint_dir, filename)
            elif name == 'model-latest':  # pragma: no cover
                path = os.path.join(self.exported_model_dir, filename)
            elif name == 'config':  # pragma: no cover
                path = os.path.join(self.config_dir, filename)
            else:
                path = os.path.join(self.base_dir, filename)

            if self.connexion.download_some_file(f["large"], object_name, path):
                logger.info('{} downloaded successfully'.format(filename))
            else:
                logger.error("Could not download {} file".format(filename))
    @exception_handler
    @beartype
    def _download_artifacts_without_tree_for_experiment(self):
        self.base_dir = self.name
        self._create_dir(self.base_dir)
        for f in self.files:
            object_name = f["object_name"]
            filename = os.path.split(f["object_name"])[-1]
            if self.connexion.download_some_file(f["large"], object_name, os.path.join(self.base_dir, filename)):
                logger.info('{} downloaded successfully'.format(filename))
            else:
                logger.error("Could not download {} file".format(filename))



from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

class Worker(Dao):

    def __init__(self, connexion: Connexion, username: str):
        super().__init__(connexion)
        self.username = username

    @exception_handler
    @beartype
    def get_infos(self) -> dict:
        """Retrieve worker info

        Examples:
            ```python
                worker = project.list_workers()[0]
                print(worker.get_infos())
            ```

        Returns:
            A dict with data of the worker
        """
        return {"username": self.username}




import json
import logging
import os
from typing import List
from beartype import beartype

from picsellia.decorators import exception_handler
from picsellia import exceptions
from picsellia.utils import generate_requirements_json, is_uuid
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset import Dataset
from picsellia.sdk.experiment import Experiment
from picsellia.sdk.model import Model
from picsellia.sdk.scan import Scan
from picsellia.sdk.worker import Worker
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger('picsellia')


class Project(Dao):

    def __init__(self, connexion: Connexion, id: str, name: str):
        super().__init__(connexion)
        self.id = id
        self.name = name

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self,) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/project/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/project/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def _convert_response_to_experiment(self, response: dict) -> Experiment:
        return Experiment(
            self.connexion,
            self.id,
            response["id"],
            response["name"],
            response["files"]
        )

    @exception_handler
    @beartype
    def list_experiments(self) -> List[Experiment]:
        """List all project's experiments

        Examples:
            ```python
                experiments = my_project.list_experiments()
            ```

        Returns:
            A list of (Experiment) objects, that you can manipulate
        """
        experiments_response = self.connexion.get(
            '/sdk/v1/project/{}/experiments'.format(self.id)).json()
        return list(map(self._convert_response_to_experiment, experiments_response["experiments"]))

    @exception_handler
    @beartype
    def delete_all_experiments(self,) -> None:
        """Delete all experiments of this project

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
                my_project.delete_all_experiments()
            ```
        """
        self.connexion.delete('/sdk/v1/project/{}/experiments'.format(self.id))
        logger.info("All experiment of project {} deleted.".format(self.name))

    @exception_handler
    @beartype
    def create_experiment(self, name: str = None, description: str = '', previous: Experiment = None,
                          dataset: Dataset = None, source: Model = None, with_artifacts: bool = False,
                          with_logs: bool = False) -> Experiment:
        """Create an experiment in this project.

        You have the same options as when creating experiments from the UI.
            - You can attach a dataset
            - You can fork a Model (it will automatically attach its files and parameters
                to the experiment)
            - You can start from a previous experiment (it will automatically attach its files and parameters
                to the new experiment)

        Examples:
            ```python
                dataset = client.get_dataset("COCO", "first_version")
                base_model = client.get_model("picsellia/yolov5")
                my_experiment = my_project.create_experiment(
                    name="new_experiment",
                    dataset=dataset,
                    source=base_model,
                )
            ```
        Arguments:
            name (str, optional): Name of experiment. Defaults to None.
            description (str, optional): Description of experiment. Defaults to ''.
            previous ((Experiment), optional): Previous experiment, if you want to base the new one on it.
                                             Defaults to None.
            dataset ((Dataset), optional): Dataset to attach. Defaults to None.
            source ((Model), optional): Model to use as source. Defaults to None.
            with_artifacts (bool, optional): Set true if you want to add files from the previous experiment.
                                             Defaults to False.
            with_logs (bool, optional): Set true if you want to add datas from the previous experiment.
                                        Defaults to False.

        Returns:
             A new (Experiment) of this project
        """
        if dataset is not None:
            dataset_id = dataset.id
        else:
            dataset_id = None
        if previous is not None:
            experiment_id = previous.id
        else:
            experiment_id = None
        if source is not None:
            model_id = source.id
        else:
            model_id = None
        data = json.dumps({
            "name": name,
            "description": description,
            "previous": experiment_id,
            "dataset": dataset_id,
            "source": model_id,
            "with_artifacts": with_artifacts,
            "with_logs": with_logs
        })
        experiment_response = self.connexion.post(
            '/sdk/v1/project/{}/experiment'.format(self.id), data=data).json()
        logger.info("Experiment {} created".format(name))
        return self._convert_response_to_experiment(experiment_response["experiment"])

    @exception_handler
    @beartype
    def update(self, **kwargs) -> None:
        """Update a project

        Examples:
            ```python
                my_project.update(description="This is a cool project")
            ```
        """
        self.connexion.patch('/sdk/v1/project/{}'.format(self.id), data=json.dumps(kwargs)).json()
        logger.info("Project {} updated.".format(self.name))

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete a project.

        :warning: **DANGER ZONE**: Be very careful here!

        It will delete the project and all experiments linked.

        Examples:
            ```python
                my_project.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/project/{}'.format(self.id))
        logger.info("Project {} deleted.".format(self.name))

    @exception_handler
    @beartype
    def get_experiment(self, name: str = None, id: str = None, tree: bool = False,
                       with_artifacts: bool = False, with_logs: bool = False) -> Experiment:
        """Retrieve an existing experiment.

        You can also to download attached files and setup according project folder tree.

        You must specify either the Experiment's name or its id.

        Examples:
            ```python
                my_experiment = my_project.get_experiment(name="test_experiment")
                my_experiment = my_project.get_experiment(id="superid", with_artifacts=True, with_logs=True)
            ```
        Arguments:
            name (str, optional): Experiment's name. Defaults to None.
            id (str, optional): Experiment's id. Defaults to None.
            tree (bool, optional): Whether to create folder tree or not. Defaults to False.
            with_artifacts (bool, optional): Whether to download every experiment's Artifacts or not. Defaults to False.
            with_logs (bool, optional): Whether to retrieve experiment's Logs or not. Defaults to False.

        Raises:
            Exception: Experiment not found

        Returns:
            An (Experiment) object that you can manipulate
        """
        identifier = None
        if id is not None:
            identifier = id
        elif name is not None:
            identifier = name
        else:
            raise exceptions.PicselliaError('No corresponding experiment found, \
                                             please enter an experiment id or an experiment name')
        experiment = self._get_experiment(identifier=identifier, with_artifacts=with_artifacts, with_logs=with_logs)

        experiment._download_artifacts_and_prepare_tree(tree, with_artifacts)

        return experiment

    @exception_handler
    @beartype
    def _get_experiment(self, identifier: str,
                        with_artifacts: bool = False, with_logs: bool = False) -> Experiment:
        data = {
            'with_artifacts': with_artifacts,
            'with_logs': with_logs
        }
        if is_uuid(identifier):
            r = self.connexion.get(
                '/sdk/v1/experiment/{}'.format(identifier), data).json()
        else:
            r = self.connexion.get(
                '/sdk/v1/project/{}/experiment?name={}'.format(self.id, identifier), data).json()
        experiment = self._convert_response_to_experiment(r["experiment"])
        return experiment

    @exception_handler
    @beartype
    def attach_dataset(self, dataset: Dataset) -> None:
        """Attach a dataset to this project.

        Retrieve or create a dataset and attach it to this project.

        Examples:
            ```python
                foo_dataset = client.create_dataset("foo", "first")
                my_project.attach_dataset(foo_dataset)
            ```
        Arguments:
            dataset (Dataset): A dataset to attach to the project.
        """
        self.connexion.post('/sdk/v1/project/{}/attach_dataset/{}'.format(self.id, dataset.id))
        logger.info("Dataset {}/{} successfully attached to project {}"
                    .format(dataset.name, dataset.version, self.name))

    @exception_handler
    @beartype
    def _convert_response_to_worker(self, response) -> Worker:
        return Worker(
            self.connexion,
            response["user"]["username"]
        )

    @exception_handler
    @beartype
    def list_workers(self,) -> List[Worker]:
        """List workers of this project.

        List all collaborators working on this project.

        Examples:
            ```python
                workers = my_project.list_workers()
                print([worker.get_infos() for worker in workers])
            ```

        Returns:
            A list of (Worker) that you can manipulate :woman_technologist: :man_technologist:
        """
        r = self.connexion.get('/sdk/v1/project/{}/workers'.format(self.id)).json()
        return list(map(self._convert_response_to_worker, r["workers"]))

    @exception_handler
    @beartype
    def _convert_response_to_scan(self, response: dict) -> Scan:
        return Scan(
            self.connexion,
            self.id,
            response["id"],
            response["name"],
        )

    @exception_handler
    @beartype
    def create_scan(self, name: str, config: dict, nb_worker: int = 1) -> Scan:
        """Initialize a new scan.

        See full documentation https://docs.picsellia.com/experiments/hyperparameter-tuning/config

        Arguments:
            name (str): Scan's name
            config (dict): config dictionnary
            nb_worker (int, optional): Number of worker to instantiate (if running remote). Defaults to 1.

        Returns:
            A (Scan) object that you can manipulate
        """
        if "script" in config.keys():
            path = config["script"]
            filename = os.path.split(path)[-1]
            object_name = os.path.join(self.id, filename)
            response = self.connexion.upload_file(path, object_name)
        else:
            object_name = None
            filename = None

        if "requirements" in config.keys():
            requirements = config["requirements"]
            if isinstance(requirements, str):
                j = generate_requirements_json(requirements)
                config["requirements"] = j["requirements"]

            elif isinstance(requirements, list):
                for e in requirements:
                    assert isinstance(
                        e, dict), "Requirements must be a list of dict"
                    assert "package" in e.keys(), "The dictionnaries must contain the key package"
                    assert "version" in e.keys(), "The dictionnaries must contain the key version"

            else:
                raise exceptions.InvalidQueryError(
                    "Please remove the key requirements from config dict if you don't want to specify any requirements")
        if "data" in config:
            data_list = config["data"]
            files = []
            assert isinstance(
                data_list, list), "data must be a list of filenames"

            for path in data_list:
                filename = os.path.split(path)[-1]
                object_name = os.path.join(self.id, filename)
                response = self.connexion.upload_file(path, object_name)
                if response.status_code == 204:
                    files.append({"filename": filename, "object_name": object_name})
        else:
            files = None
        if files is not None:
            config["data"] = files
        data = json.dumps({
            "name": name,
            "config": config,
            "nb_worker": nb_worker,
            "object_name": object_name,
            "filename": filename,
        })
        r = self.connexion.post(
            '/sdk/v1/project/{}/scan'.format(self.id), data=data).json()
        scan = self._convert_response_to_scan(r["scan"])
        return scan




from picsellia.types.prediction_format import PredictionFormat


class ClassificationPredictionFormat(PredictionFormat):

    def __init__(self, label_id : int, score : float):
        self.label_id = label_id
        self.score = float(score)
        self.check_validity()

    def check_validity(self) -> bool:
        assert isinstance(self.label_id, int), "Label of classification prediction shall be id in labelmap and should be an int"
        assert isinstance(self.score, float), "Score of classification prediction shall be a float"
        assert self.score >= 0, "Score shall be positive"

    def to_payload(self) -> dict:
        self.check_validity()
        return {
            "labels" : [self.label_id],
            "detection_scores" : [self.score]
        }